#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

import os
from typing import Any, Iterable, List, Optional, Type, Union, cast, overload

from pydantic import BaseModel, ValidationError

from neo4j_graphrag.exceptions import LLMGenerationError
from neo4j_graphrag.llm.base import LLMInterface, LLMInterfaceV2
from neo4j_graphrag.llm.types import (
    BaseMessage,
    LLMResponse,
    MessageList,
    SystemMessage,
    UserMessage,
)
from neo4j_graphrag.message_history import MessageHistory
from neo4j_graphrag.types import LLMMessage
from neo4j_graphrag.utils.rate_limit import (
    RateLimitHandler,
)
from neo4j_graphrag.utils.rate_limit import (
    async_rate_limit_handler as async_rate_limit_handler_decorator,
)
from neo4j_graphrag.utils.rate_limit import (
    rate_limit_handler as rate_limit_handler_decorator,
)

try:
    from mistralai import (
        AssistantMessage,
        Messages,
        Mistral,
    )
    from mistralai import (
        SystemMessage as MistralSystemMessage,
    )
    from mistralai import (
        UserMessage as MistralUserMessage,
    )
    from mistralai.models.sdkerror import SDKError
except ImportError:
    Mistral = None  # type: ignore
    SDKError = None  # type: ignore


# pylint: disable=redefined-builtin, arguments-differ, raise-missing-from, no-else-return
class MistralAILLM(LLMInterface, LLMInterfaceV2):
    def __init__(
        self,
        model_name: str,
        model_params: Optional[dict[str, Any]] = None,
        rate_limit_handler: Optional[RateLimitHandler] = None,
        **kwargs: Any,
    ):
        """

        Args:
            model_name (str):
            model_params (str): Parameters for LLMInterface(V1) like temperature and such that will be
             passed to the chat completions endpoint
            rate_limit_handler (Optional[RateLimitHandler]): Handler for rate limiting for LLMInterface(V1). Defaults to retry with exponential backoff.
            kwargs: All other parameters will be passed to the Mistral client.

        """
        if Mistral is None:
            raise ImportError(
                """Could not import Mistral Python client.
                Please install it with `pip install "neo4j-graphrag[mistralai]"`."""
            )
        LLMInterfaceV2.__init__(
            self,
            model_name=model_name,
            model_params=model_params or {},
            rate_limit_handler=rate_limit_handler,
            **kwargs,
        )
        api_key = kwargs.pop("api_key", None)
        if api_key is None:
            api_key = os.getenv("MISTRAL_API_KEY", "")
        self.client = Mistral(api_key=api_key, **kwargs)

    # overloads for LLMInterface and LLMInterfaceV2 methods
    @overload  # type: ignore[no-overload-impl]
    def invoke(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse: ...

    @overload
    def invoke(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse: ...

    @overload  # type: ignore[no-overload-impl]
    async def ainvoke(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse: ...

    @overload
    async def ainvoke(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse: ...

    # switching logics to LLMInterface or LLMInterfaceV2
    def invoke(  # type: ignore[no-redef]
        self,
        input: Union[str, List[LLMMessage]],
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        if isinstance(input, str):
            return self.__invoke_v1(input, message_history, system_instruction)
        elif isinstance(input, list):
            return self.__invoke_v2(input, response_format=response_format, **kwargs)
        else:
            raise ValueError(f"Invalid input type for invoke method - {type(input)}")

    async def ainvoke(  # type: ignore[no-redef]
        self,
        input: Union[str, List[LLMMessage]],
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        if isinstance(input, str):
            return await self.__ainvoke_v1(input, message_history, system_instruction)
        elif isinstance(input, list):
            return await self.__ainvoke_v2(
                input, response_format=response_format, **kwargs
            )
        else:
            raise ValueError(f"Invalid input type for ainvoke method - {type(input)}")

    # implementations
    @rate_limit_handler_decorator
    def __invoke_v1(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse:
        """Sends a text input to the Mistral chat completion model
        and returns the response's content.

        Args:
            input (str): Text sent to the LLM.
            message_history (Optional[Union[List[LLMMessage], MessageHistory]]): A collection previous messages, with each message having a specific role assigned.
            system_instruction (Optional[str]): An option to override the llm system message for this invocation.

        Returns:
            LLMResponse: The response from MistralAI.

        Raises:
            LLMGenerationError: If anything goes wrong.
        """
        try:
            if isinstance(message_history, MessageHistory):
                message_history = message_history.messages
            messages = self.get_messages(input, message_history, system_instruction)
            response = self.client.chat.complete(
                model=self.model_name,
                messages=messages,
                **self.model_params,
            )
            content: str = ""
            if response and response.choices:
                possible_content = response.choices[0].message.content
                if isinstance(possible_content, str):
                    content = possible_content
            return LLMResponse(content=content)
        except SDKError as e:
            raise LLMGenerationError(e)

    def __invoke_v2(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        """Sends a text input to the Mistral chat completion model
        and returns the response's content.

        Args:
            input (List[LLMMessage]): Messages sent to the LLM.
            response_format: Not supported by MistralAILLM.

        Returns:
            LLMResponse: The response from MistralAI.

        Raises:
            LLMGenerationError: If anything goes wrong.
        """
        if response_format is not None:
            raise NotImplementedError(
                "MistralAILLM does not currently support structured output"
            )
        try:
            messages = self.get_messages_v2(input)
            response = self.client.chat.complete(
                model=self.model_name, messages=messages, **self.model_params, **kwargs
            )
            content: str = ""
            if response and response.choices:
                possible_content = response.choices[0].message.content
                if isinstance(possible_content, str):
                    content = possible_content
            return LLMResponse(content=content)
        except SDKError as e:
            raise LLMGenerationError(e)

    @async_rate_limit_handler_decorator
    async def __ainvoke_v1(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse:
        """Asynchronously sends a text input to the MistralAI chat
        completion model and returns the response's content.

        Args:
            input (str): Text sent to the LLM.
            message_history (Optional[Union[List[LLMMessage], MessageHistory]]): A collection previous messages,
                with each message having a specific role assigned.
            system_instruction (Optional[str]): An option to override the llm system message for this invocation.

        Returns:
            LLMResponse: The response from MistralAI.

        Raises:
            LLMGenerationError: If anything goes wrong.
        """
        try:
            if isinstance(message_history, MessageHistory):
                message_history = message_history.messages
            messages = self.get_messages(input, message_history, system_instruction)
            response = await self.client.chat.complete_async(
                model=self.model_name,
                messages=messages,
                **self.model_params,
            )
            content: str = ""
            if response and response.choices:
                possible_content = response.choices[0].message.content
                if isinstance(possible_content, str):
                    content = possible_content
            return LLMResponse(content=content)
        except SDKError as e:
            raise LLMGenerationError(e)

    async def __ainvoke_v2(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        """Asynchronously sends a text input to the MistralAI chat
        completion model and returns the response's content.

        Args:
            input (List[LLMMessage]): Messages sent to the LLM.
            response_format: Not supported by MistralAILLM.

        Returns:
            LLMResponse: The response from MistralAI.

        Raises:
            LLMGenerationError: If anything goes wrong.
        """
        if response_format is not None:
            raise NotImplementedError(
                "MistralAILLM does not currently support structured output"
            )
        try:
            messages = self.get_messages_v2(input)
            response = await self.client.chat.complete_async(
                model=self.model_name,
                messages=messages,
                **self.model_params,
                **kwargs,
            )
            content: str = ""
            if response and response.choices:
                possible_content = response.choices[0].message.content
                if isinstance(possible_content, str):
                    content = possible_content
            return LLMResponse(content=content)
        except SDKError as e:
            raise LLMGenerationError(e)

    # subsidiary methods
    def get_messages(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> list[Messages]:
        """Constructs the message list for the Mistral chat completion model."""
        messages = []
        if system_instruction:
            messages.append(SystemMessage(content=system_instruction).model_dump())
        if message_history:
            if isinstance(message_history, MessageHistory):
                message_history = message_history.messages
            try:
                MessageList(messages=cast(list[BaseMessage], message_history))
            except ValidationError as e:
                raise LLMGenerationError(e.errors()) from e
            messages.extend(cast(Iterable[dict[str, Any]], message_history))
        messages.append(UserMessage(content=input).model_dump())
        return cast(list[Messages], messages)

    def get_messages_v2(
        self,
        input: list[LLMMessage],
    ) -> list[Messages]:
        """Constructs the message list for the Mistral chat completion model."""
        messages: list[Messages] = []
        for m in input:
            if m["role"] == "system":
                messages.append(MistralSystemMessage(content=m["content"]))
                continue
            if m["role"] == "user":
                messages.append(MistralUserMessage(content=m["content"]))
                continue
            if m["role"] == "assistant":
                messages.append(AssistantMessage(content=m["content"]))
                continue
            raise ValueError(f"Unknown role: {m['role']}")
        return messages



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# built-in dependencies
from __future__ import annotations

import warnings
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    List,
    Optional,
    Sequence,
    Type,
    Union,
    cast,
    Dict,
    overload,
)

# 3rd-party dependencies
from pydantic import BaseModel, ValidationError

# project dependencies
from neo4j_graphrag.exceptions import LLMGenerationError
from neo4j_graphrag.message_history import MessageHistory
from neo4j_graphrag.types import LLMMessage
from neo4j_graphrag.utils.rate_limit import (
    RateLimitHandler,
)
from neo4j_graphrag.utils.rate_limit import (
    async_rate_limit_handler as async_rate_limit_handler_decorator,
)
from neo4j_graphrag.utils.rate_limit import (
    rate_limit_handler as rate_limit_handler_decorator,
)

from .base import LLMInterface, LLMInterfaceV2
from .types import (
    BaseMessage,
    LLMResponse,
    MessageList,
    ToolCall,
    ToolCallResponse,
    SystemMessage,
    UserMessage,
)
from neo4j_graphrag.tool import Tool

if TYPE_CHECKING:
    from ollama import Message

# pylint: disable=redefined-builtin, arguments-differ, raise-missing-from, no-else-return, import-outside-toplevel


class OllamaLLM(LLMInterface, LLMInterfaceV2):
    """LLM wrapper for Ollama models."""

    def __init__(
        self,
        model_name: str,
        model_params: Optional[dict[str, Any]] = None,
        rate_limit_handler: Optional[RateLimitHandler] = None,
        **kwargs: Any,
    ):
        try:
            import ollama
        except ImportError:
            raise ImportError(
                "Could not import ollama Python client. "
                "Please install it with `pip install ollama`."
            )
        LLMInterfaceV2.__init__(
            self,
            model_name=model_name,
            model_params=model_params or {},
            rate_limit_handler=rate_limit_handler,
            **kwargs,
        )
        self.ollama = ollama
        self.client = ollama.Client(
            **kwargs,
        )
        self.async_client = ollama.AsyncClient(
            **kwargs,
        )
        if "stream" in self.model_params:
            raise ValueError("Streaming is not supported by the OllamaLLM wrapper")
        # bug-fix with backward compatibility:
        # we mistakenly passed all "model_params" under the options argument
        # next two lines to be removed in 2.0
        if not any(
            key in self.model_params for key in ("options", "format", "keep_alive")
        ):
            warnings.warn(
                """Passing options directly without including them in an 'options' key is deprecated. Ie you must use model_params={"options": {"temperature": 0}}""",
                DeprecationWarning,
            )
            self.model_params = {"options": self.model_params}

    # overloads for LLMInterface and LLMInterfaceV2 methods
    @overload  # type: ignore[no-overload-impl]
    def invoke(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse: ...

    @overload
    def invoke(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse: ...

    @overload  # type: ignore[no-overload-impl]
    async def ainvoke(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse: ...

    @overload
    async def ainvoke(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse: ...

    # switching logics to LLMInterface or LLMInterfaceV2
    def invoke(  # type: ignore[no-redef]
        self,
        input: Union[str, List[LLMMessage]],
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        if isinstance(input, str):
            return self.__invoke_v1(input, message_history, system_instruction)
        elif isinstance(input, list):
            return self.__invoke_v2(input, response_format=response_format, **kwargs)
        else:
            raise ValueError(f"Invalid input type for invoke method - {type(input)}")

    async def ainvoke(  # type: ignore[no-redef]
        self,
        input: Union[str, List[LLMMessage]],
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        if isinstance(input, str):
            return await self.__ainvoke_v1(input, message_history, system_instruction)
        elif isinstance(input, list):
            return await self.__ainvoke_v2(
                input, response_format=response_format, **kwargs
            )
        else:
            raise ValueError(f"Invalid input type for ainvoke method - {type(input)}")

    @rate_limit_handler_decorator
    def __invoke_v1(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse:
        """Sends text to the LLM and returns a response.

        Args:
            input (str): The text to send to the LLM.
            message_history (Optional[Union[List[LLMMessage], MessageHistory]]): A collection previous messages,
                with each message having a specific role assigned.
            system_instruction (Optional[str]): An option to override the llm system message for this invocation.

        Returns:
            LLMResponse: The response from the LLM.
        """
        try:
            if isinstance(message_history, MessageHistory):
                message_history = message_history.messages
            response = self.client.chat(
                model=self.model_name,
                messages=self.get_messages(input, message_history, system_instruction),
                **self.model_params,
            )
            content = response.message.content or ""
            return LLMResponse(content=content)
        except self.ollama.ResponseError as e:
            raise LLMGenerationError(e)

    @rate_limit_handler_decorator
    def __invoke_v2(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        """Sends text to the LLM and returns a response.

        Args:
            input (List[LLMMessage]): The messages to send to the LLM.
            response_format: Not supported by OllamaLLM.

        Returns:
            LLMResponse: The response from the LLM.
        """
        if response_format is not None:
            raise NotImplementedError(
                "OllamaLLM does not currently support structured output"
            )
        try:
            response = self.client.chat(
                model=self.model_name,
                messages=self.get_messages_v2(input),
                **self.model_params,
                **kwargs,
            )
            content = response.message.content or ""
            return LLMResponse(content=content)
        except self.ollama.ResponseError as e:
            raise LLMGenerationError(e)

    @async_rate_limit_handler_decorator
    async def __ainvoke_v1(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse:
        """Asynchronously sends a text input to the OpenAI chat
        completion model and returns the response's content.

        Args:
            input (str): Text sent to the LLM.
            message_history (Optional[Union[List[LLMMessage], MessageHistory]]): A collection previous messages,
                with each message having a specific role assigned.
            system_instruction (Optional[str]): An option to override the llm system message for this invocation.

        Returns:
            LLMResponse: The response from OpenAI.

        Raises:
            LLMGenerationError: If anything goes wrong.
        """
        try:
            if isinstance(message_history, MessageHistory):
                message_history = message_history.messages
            response = await self.async_client.chat(
                model=self.model_name,
                messages=self.get_messages(input, message_history, system_instruction),
                options=self.model_params,
            )
            content = response.message.content or ""
            return LLMResponse(content=content)
        except self.ollama.ResponseError as e:
            raise LLMGenerationError(e)

    @async_rate_limit_handler_decorator
    async def __ainvoke_v2(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        """Asynchronously sends a text input to the Ollama chat
        completion model and returns the response's content.

        Args:
            input (List[LLMMessage]): Messages sent to the LLM.
            response_format: Not supported by OllamaLLM.

        Returns:
            LLMResponse: The response from Ollama.

        Raises:
            LLMGenerationError: If anything goes wrong.
        """
        if response_format is not None:
            raise NotImplementedError(
                "OllamaLLM does not currently support structured output"
            )
        try:
            params = {**self.model_params, **kwargs}
            response = await self.async_client.chat(
                model=self.model_name,
                messages=self.get_messages_v2(input),
                options=params,
            )
            content = response.message.content or ""
            return LLMResponse(content=content)
        except self.ollama.ResponseError as e:
            raise LLMGenerationError(e)

    # subsdiary methods
    def get_messages(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> Sequence[Message]:
        """Constructs the message list for the Ollama chat API."""
        messages = []
        if system_instruction:
            messages.append(SystemMessage(content=system_instruction).model_dump())
        if message_history:
            if isinstance(message_history, MessageHistory):
                message_history = message_history.messages
            try:
                MessageList(messages=cast(list[BaseMessage], message_history))
            except ValidationError as e:
                raise LLMGenerationError(e.errors()) from e
            messages.extend(cast(Iterable[dict[str, Any]], message_history))
        messages.append(UserMessage(content=input).model_dump())
        return messages  # type: ignore

    def get_messages_v2(
        self,
        input: list[LLMMessage],
    ) -> Sequence[Message]:
        """Constructs the message list for the Ollama chat API."""
        return [self.ollama.Message(**i) for i in input]

    @rate_limit_handler_decorator
    def invoke_with_tools(
        self,
        input: str,
        tools: Sequence[Tool],  # Tools definition as a sequence of Tool objects
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> ToolCallResponse:
        """Sends a text input to the LLM with tool definitions
        and retrieves a tool call response.
        Args:
            input (str): Text sent to the LLM.
            tools (List[Tool]): List of Tools for the LLM to choose from.
            message_history (Optional[Union[List[LLMMessage], MessageHistory]]): A collection previous messages,
                with each message having a specific role assigned.
            system_instruction (Optional[str]): An option to override the llm system message for this invocation.
        Returns:
            ToolCallResponse: The response from the LLM containing a tool call.
        Raises:
            LLMGenerationError: If anything goes wrong.
        """
        try:
            if isinstance(message_history, MessageHistory):
                message_history = message_history.messages

            # Convert tools to Ollama's expected type
            ollama_tools = []
            for tool in tools:
                ollama_tool_format = self._convert_tool_to_ollama_format(tool)
                ollama_tools.append(ollama_tool_format)
            response = self.client.chat(
                model=self.model_name,
                messages=self.get_messages(input, message_history, system_instruction),
                tools=ollama_tools,
                **self.model_params,
            )
            message = response.message
            # If there's no tool call, return the content as a regular response
            if not message.tool_calls or len(message.tool_calls) == 0:
                return ToolCallResponse(
                    tool_calls=[],
                    content=message.content,
                )

            # Process all tool calls
            tool_calls = []

            for tool_call in message.tool_calls:
                args = tool_call.function.arguments
                tool_calls.append(
                    ToolCall(name=tool_call.function.name, arguments=args)
                )

            return ToolCallResponse(tool_calls=tool_calls, content=message.content)
        except self.ollama.ResponseError as e:
            raise LLMGenerationError(e)

    @async_rate_limit_handler_decorator
    async def ainvoke_with_tools(
        self,
        input: str,
        tools: Sequence[Tool],  # Tools definition as a sequence of Tool objects
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> ToolCallResponse:
        """Sends a text input to the LLM with tool definitions
        and retrieves a tool call response.
        Args:
            input (str): Text sent to the LLM.
            tools (List[Tool]): List of Tools for the LLM to choose from.
            message_history (Optional[Union[List[LLMMessage], MessageHistory]]): A collection previous messages,
                with each message having a specific role assigned.
            system_instruction (Optional[str]): An option to override the llm system message for this invocation.
        Returns:
            ToolCallResponse: The response from the LLM containing a tool call.
        Raises:
            LLMGenerationError: If anything goes wrong.
        """
        try:
            if isinstance(message_history, MessageHistory):
                message_history = message_history.messages

            # Convert tools to Ollama's expected type
            ollama_tools = []
            for tool in tools:
                ollama_tool_format = self._convert_tool_to_ollama_format(tool)
                ollama_tools.append(ollama_tool_format)

            response = await self.async_client.chat(
                model=self.model_name,
                messages=self.get_messages(input, message_history, system_instruction),
                tools=ollama_tools,
                **self.model_params,
            )
            message = response.message

            # If there's no tool call, return the content as a regular response
            if not message.tool_calls or len(message.tool_calls) == 0:
                return ToolCallResponse(
                    tool_calls=[],
                    content=message.content,
                )

            # Process all tool calls
            tool_calls = []

            for tool_call in message.tool_calls:
                args = tool_call.function.arguments
                tool_calls.append(
                    ToolCall(name=tool_call.function.name, arguments=args)
                )

            return ToolCallResponse(tool_calls=tool_calls, content=message.content)
        except self.ollama.ResponseError as e:
            raise LLMGenerationError(e)

    def _convert_tool_to_ollama_format(self, tool: Tool) -> Dict[str, Any]:
        """Convert a Tool object to Ollama's expected format.
        Args:
            tool: A Tool object to convert to Ollama's format.
        Returns:
            A dictionary in Ollama's tool format.
        """
        try:
            return {
                "type": "function",
                "function": {
                    "name": tool.get_name(),
                    "description": tool.get_description(),
                    "parameters": tool.get_parameters(),
                },
            }
        except AttributeError:
            raise LLMGenerationError(f"Tool {tool} is not a valid Tool object")



"""In this example, the pipeline is defined in a JSON ('simple_kg_pipeline_config.json')
or YAML ('simple_kg_pipeline_config.yaml') file.

According to the configuration file, some parameters will be read from the env vars
(Neo4j credentials and the OpenAI API key).
"""

import asyncio
import logging

## If env vars are in a .env file, uncomment:
## (requires pip install python-dotenv)
# from dotenv import load_dotenv
# load_dotenv()
# env vars manually set for testing:
import os
from pathlib import Path

from neo4j_graphrag.experimental.pipeline.config.runner import PipelineRunner
from neo4j_graphrag.experimental.pipeline.pipeline import PipelineResult

logging.basicConfig()
logging.getLogger("neo4j_graphrag").setLevel(logging.DEBUG)

os.environ["NEO4J_URI"] = "bolt://localhost:7687"
os.environ["NEO4J_USER"] = "neo4j"
os.environ["NEO4J_PASSWORD"] = "password"
# os.environ["OPENAI_API_KEY"] = "sk-..."


root_dir = Path(__file__).parent
file_path = root_dir / "simple_kg_pipeline_config_url.json"


# File to process
URL = "https://raw.githubusercontent.com/neo4j/neo4j-graphrag-python/c166afc4d5abc56a5686f3da46a97ed7c07da19d/examples/data/Harry%20Potter%20and%20the%20Chamber%20of%20Secrets%20Summary.pdf"


async def main() -> PipelineResult:
    pipeline = PipelineRunner.from_config_file(file_path)
    return await pipeline.run({"file_path": URL})


if __name__ == "__main__":
    print(asyncio.run(main()))



from unittest.mock import MagicMock, Mock, patch

import numpy as np
import pytest
import torch
from neo4j_graphrag.embeddings.base import Embedder
from neo4j_graphrag.embeddings.sentence_transformers import (
    SentenceTransformerEmbeddings,
)
from neo4j_graphrag.exceptions import EmbeddingsGenerationError


def get_mock_sentence_transformers() -> MagicMock:
    mock = MagicMock()
    # I know, I know... ¯\_(ツ)_/¯
    # This is to cover the if type checks in the embed_query method
    mock.Tensor = torch.Tensor
    mock.ndarray = np.ndarray
    return mock


@patch("builtins.__import__")
def test_initialization(mock_import: Mock) -> None:
    MockSentenceTransformer = get_mock_sentence_transformers()
    mock_import.return_value = MockSentenceTransformer
    instance = SentenceTransformerEmbeddings()
    MockSentenceTransformer.SentenceTransformer.assert_called_with("all-MiniLM-L6-v2")
    assert isinstance(instance, Embedder)


@patch("builtins.__import__")
def test_initialization_with_custom_model(mock_import: Mock) -> None:
    MockSentenceTransformer = get_mock_sentence_transformers()
    mock_import.return_value = MockSentenceTransformer
    custom_model = "distilbert-base-nli-stsb-mean-tokens"
    SentenceTransformerEmbeddings(model=custom_model)
    MockSentenceTransformer.SentenceTransformer.assert_called_with(custom_model)


@patch("builtins.__import__")
def test_embed_query(mock_import: Mock) -> None:
    MockSentenceTransformer = get_mock_sentence_transformers()
    mock_import.return_value = MockSentenceTransformer
    mock_model = MockSentenceTransformer.SentenceTransformer.return_value
    mock_model.encode.return_value = np.array([[0.1, 0.2, 0.3]])

    instance = SentenceTransformerEmbeddings()
    result = instance.embed_query("test query")

    mock_model.encode.assert_called_with(["test query"])
    assert isinstance(result, list)
    assert result == [0.1, 0.2, 0.3]


@patch("builtins.__import__", side_effect=ImportError)
def test_import_error(mock_import: Mock) -> None:
    with pytest.raises(ImportError):
        SentenceTransformerEmbeddings()


@patch("builtins.__import__")
def test_embed_query_non_retryable_error_handling(mock_import: Mock) -> None:
    """Test that non-retryable errors fail immediately without retries."""
    MockSentenceTransformer = get_mock_sentence_transformers()
    mock_import.return_value = MockSentenceTransformer
    mock_model = MockSentenceTransformer.SentenceTransformer.return_value
    mock_model.encode.side_effect = Exception("Model error")

    instance = SentenceTransformerEmbeddings()
    with pytest.raises(
        EmbeddingsGenerationError,
        match="Failed to generate embedding with SentenceTransformer",
    ):
        instance.embed_query("test query")

    # Verify the model was called only once (no retries for non-rate-limit errors)
    assert mock_model.encode.call_count == 1



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from pydantic import BaseModel, ConfigDict
from collections.abc import Awaitable

from typing import Any, Optional, Protocol, runtime_checkable


@runtime_checkable
class TaskProgressNotifierProtocol(Protocol):
    """This protocol is used to send events from the component to the
    Pipeline callback protocol.
    The event sent to the callback will be of type :ref:`TaskEvent`,
    with `event_type=TASK_PROGRESS`.
    """

    def __call__(self, message: str, data: dict[str, Any]) -> Awaitable[None]: ...


class RunContext(BaseModel):
    """Context passed to the component"""

    run_id: str
    task_name: str
    notifier: Optional[TaskProgressNotifierProtocol] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    async def notify(self, message: str, data: dict[str, Any]) -> None:
        if self.notifier:
            await self.notifier(message=message, data=data)



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

from typing import Any

import neo4j
from neo4j import GraphDatabase
from qdrant_client import QdrantClient, models

from ..utils import build_data_objects, populate_neo4j


def populate_dbs(
    neo4j_driver: neo4j.Driver, client: QdrantClient, collection_name: str = "Jeopardy"
) -> None:
    neo4j_objects, question_objs = build_data_objects("qdrant")

    if client.collection_exists(collection_name):
        client.delete_collection(collection_name)

    client.create_collection(
        collection_name=collection_name,
        vectors_config=models.VectorParams(size=384, distance=models.Distance.COSINE),
    )

    populate_qdrant(client, question_objs, collection_name)

    populate_neo4j(neo4j_driver, neo4j_objects)


def populate_qdrant(
    client: QdrantClient, question_objs: list[Any], collection_name: str
) -> None:
    client.upsert(collection_name=collection_name, points=question_objs)


if __name__ == "__main__":
    NEO4J_AUTH = ("neo4j", "password")
    NEO4J_URL = "neo4j://localhost:7687"
    with GraphDatabase.driver(NEO4J_URL, auth=NEO4J_AUTH) as neo4j_driver:
        populate_dbs(neo4j_driver, QdrantClient(url="http://localhost:6333"))



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

from typing import Any
from unittest import mock
from unittest.mock import MagicMock, Mock

import pytest
from neo4j_graphrag.experimental.components.kg_writer import Neo4jWriter, batched
from neo4j_graphrag.experimental.components.types import (
    LexicalGraphConfig,
    Neo4jGraph,
    Neo4jNode,
    Neo4jRelationship,
)
from neo4j_graphrag.neo4j_queries import (
    upsert_node_query,
    upsert_relationship_query,
)


def test_batched() -> None:
    assert list(batched([1, 2, 3, 4], batch_size=2)) == [
        [1, 2],
        [3, 4],
    ]
    assert list(batched([1, 2, 3], batch_size=2)) == [
        [1, 2],
        [3],
    ]
    assert list(batched([1, 2, 3], batch_size=4)) == [
        [1, 2, 3],
    ]


@mock.patch(
    "neo4j_graphrag.experimental.components.kg_writer.get_version",
    return_value=((5, 22, 0), False, False),
)
@mock.patch(
    "neo4j_graphrag.experimental.components.kg_writer.Neo4jWriter._db_setup",
    return_value=None,
)
def test_upsert_nodes(_: Mock, driver: MagicMock) -> None:
    driver.execute_query.return_value = (
        [{"element_id": "#1"}],
        None,
        None,
    )
    neo4j_writer = Neo4jWriter(driver=driver)
    node = Neo4jNode(id="1", label="Label", properties={"key": "value"})
    neo4j_writer._upsert_nodes(nodes=[node], lexical_graph_config=LexicalGraphConfig())
    driver.execute_query.assert_called_once_with(
        upsert_node_query(False),
        parameters_={
            "rows": [
                {
                    "label": "Label",
                    "labels": ["Label", "__Entity__"],
                    "id": "1",
                    "properties": {"key": "value"},
                    "embedding_properties": {},
                }
            ]
        },
        database_=None,
    )


@mock.patch(
    "neo4j_graphrag.experimental.components.kg_writer.get_version",
    return_value=((5, 22, 0), False, False),
)
@mock.patch(
    "neo4j_graphrag.experimental.components.kg_writer.Neo4jWriter._db_setup",
    return_value=None,
)
def test_upsert_nodes_with_embedding(
    _: Mock,
    driver: MagicMock,
) -> None:
    driver.execute_query.return_value = (
        [{"element_id": "#1"}],
        None,
        None,
    )
    neo4j_writer = Neo4jWriter(driver=driver)
    node = Neo4jNode(
        id="1",
        label="Label",
        properties={"key": "value"},
        embedding_properties={"embeddingProp": [1.0, 2.0, 3.0]},
    )
    neo4j_writer._upsert_nodes(nodes=[node], lexical_graph_config=LexicalGraphConfig())
    driver.execute_query.assert_any_call(
        upsert_node_query(False),
        parameters_={
            "rows": [
                {
                    "label": "Label",
                    "labels": ["Label", "__Entity__"],
                    "id": "1",
                    "properties": {"key": "value"},
                    "embedding_properties": {"embeddingProp": [1.0, 2.0, 3.0]},
                }
            ]
        },
        database_=None,
    )


@mock.patch(
    "neo4j_graphrag.experimental.components.kg_writer.get_version",
    return_value=((5, 22, 0), False, False),
)
@mock.patch(
    "neo4j_graphrag.experimental.components.kg_writer.Neo4jWriter._db_setup",
    return_value=None,
)
def test_upsert_relationship(_: Mock, driver: MagicMock) -> None:
    neo4j_writer = Neo4jWriter(driver=driver)
    rel = Neo4jRelationship(
        start_node_id="1",
        end_node_id="2",
        type="RELATIONSHIP",
        properties={"key": "value"},
    )
    neo4j_writer._upsert_relationships(
        rels=[rel],
    )
    parameters = {
        "rows": [
            {
                "type": "RELATIONSHIP",
                "start_node_id": "1",
                "end_node_id": "2",
                "properties": {"key": "value"},
                "embedding_properties": {},
            }
        ]
    }
    driver.execute_query.assert_called_once_with(
        upsert_relationship_query(False),
        parameters_=parameters,
        database_=None,
    )


@mock.patch(
    "neo4j_graphrag.experimental.components.kg_writer.get_version",
    return_value=((5, 22, 0), False, False),
)
@mock.patch(
    "neo4j_graphrag.experimental.components.kg_writer.Neo4jWriter._db_setup",
    return_value=None,
)
def test_upsert_relationship_with_embedding(_: Mock, driver: MagicMock) -> None:
    neo4j_writer = Neo4jWriter(driver=driver)
    rel = Neo4jRelationship(
        start_node_id="1",
        end_node_id="2",
        type="RELATIONSHIP",
        properties={"key": "value"},
        embedding_properties={"embeddingProp": [1.0, 2.0, 3.0]},
    )
    driver.execute_query.return_value.records = [{"elementId(r)": "rel_elem_id"}]
    neo4j_writer._upsert_relationships(
        rels=[rel],
    )
    parameters = {
        "rows": [
            {
                "type": "RELATIONSHIP",
                "start_node_id": "1",
                "end_node_id": "2",
                "properties": {"key": "value"},
                "embedding_properties": {"embeddingProp": [1.0, 2.0, 3.0]},
            }
        ]
    }
    driver.execute_query.assert_any_call(
        upsert_relationship_query(False),
        parameters_=parameters,
        database_=None,
    )


@mock.patch(
    "neo4j_graphrag.experimental.components.kg_writer.get_version",
    return_value=((5, 22, 0), False, False),
)
@pytest.mark.asyncio
@mock.patch(
    "neo4j_graphrag.experimental.components.kg_writer.Neo4jWriter._db_setup",
    return_value=None,
)
async def test_run(_: Mock, driver: MagicMock) -> None:
    driver.execute_query.return_value = (
        [
            {"element_id": "#1"},
            {"element_id": "#2"},
        ],
        None,
        None,
    )
    neo4j_writer = Neo4jWriter(driver=driver)
    node = Neo4jNode(id="1", label="Label")
    rel = Neo4jRelationship(start_node_id="1", end_node_id="2", type="RELATIONSHIP")
    graph = Neo4jGraph(nodes=[node], relationships=[rel])
    await neo4j_writer.run(graph=graph)
    driver.execute_query.assert_any_call(
        upsert_node_query(False),
        parameters_={
            "rows": [
                {
                    "label": "Label",
                    "labels": ["Label", "__Entity__"],
                    "id": "1",
                    "properties": {},
                    "embedding_properties": {},
                }
            ]
        },
        database_=None,
    )
    parameters_: dict[str, Any] = {
        "rows": [
            {
                "type": "RELATIONSHIP",
                "start_node_id": "1",
                "end_node_id": "2",
                "properties": {},
                "embedding_properties": {},
            }
        ]
    }
    driver.execute_query.assert_any_call(
        upsert_relationship_query(False),
        parameters_=parameters_,
        database_=None,
    )


@pytest.mark.asyncio
@mock.patch(
    "neo4j_graphrag.experimental.components.kg_writer.Neo4jWriter._db_setup",
    return_value=None,
)
async def test_run_is_version_below_5_23(_: Mock) -> None:
    driver = MagicMock()
    driver.execute_query = Mock(
        side_effect=(
            # get_version
            ([{"versions": ["5.22.0"], "edition": "enterpise"}], None, None),
            # upsert nodes
            ([{"_internal_id": "1", "element_id": "#1"}], None, None),
            # upsert relationships
            (None, None, None),
        )
    )

    neo4j_writer = Neo4jWriter(driver=driver)

    node = Neo4jNode(id="1", label="Label")
    rel = Neo4jRelationship(start_node_id="1", end_node_id="2", type="RELATIONSHIP")
    graph = Neo4jGraph(nodes=[node], relationships=[rel])
    await neo4j_writer.run(graph=graph)

    driver.execute_query.assert_any_call(
        upsert_node_query(False),
        parameters_={
            "rows": [
                {
                    "label": "Label",
                    "labels": ["Label", "__Entity__"],
                    "id": "1",
                    "properties": {},
                    "embedding_properties": {},
                }
            ]
        },
        database_=None,
    )
    parameters_: dict[str, Any] = {
        "rows": [
            {
                "type": "RELATIONSHIP",
                "start_node_id": "1",
                "end_node_id": "2",
                "properties": {},
                "embedding_properties": {},
            }
        ]
    }
    driver.execute_query.assert_any_call(
        upsert_relationship_query(False),
        parameters_=parameters_,
        database_=None,
    )


@pytest.mark.asyncio
@mock.patch(
    "neo4j_graphrag.experimental.components.kg_writer.Neo4jWriter._db_setup",
    return_value=None,
)
async def test_run_is_version_5_23_or_above(_: Mock) -> None:
    driver = MagicMock()
    driver.execute_query = Mock(
        side_effect=(
            # get_version
            ([{"versions": ["5.23.0"], "edition": "enterpise"}], None, None),
            # upsert nodes
            ([{"element_id": "#1"}], None, None),
            # upsert relationships
            (None, None, None),
        )
    )

    neo4j_writer = Neo4jWriter(driver=driver)
    neo4j_writer.is_version_5_23_or_above = True

    node = Neo4jNode(id="1", label="Label")
    rel = Neo4jRelationship(start_node_id="1", end_node_id="2", type="RELATIONSHIP")
    graph = Neo4jGraph(nodes=[node], relationships=[rel])
    await neo4j_writer.run(graph=graph)

    driver.execute_query.assert_any_call(
        upsert_node_query(True),
        parameters_={
            "rows": [
                {
                    "label": "Label",
                    "labels": ["Label", "__Entity__"],
                    "id": "1",
                    "properties": {},
                    "embedding_properties": {},
                }
            ]
        },
        database_=None,
    )
    parameters_: dict[str, Any] = {
        "rows": [
            {
                "type": "RELATIONSHIP",
                "start_node_id": "1",
                "end_node_id": "2",
                "properties": {},
                "embedding_properties": {},
            }
        ]
    }
    driver.execute_query.assert_any_call(
        upsert_relationship_query(True),
        parameters_=parameters_,
        database_=None,
    )


@pytest.mark.parametrize(
    "description, version, is_5_23_or_above",
    [
        ("SemVer, < 5.23", "5.22.0", False),
        ("SemVer, > 5.23", "5.24.0", True),
        ("SemVer, < 5.23, Aura", "5.22-aura", False),
        ("SemVer, > 5.23, Aura", "5.24-aura", True),
        ("CalVer", "2025.01.0", True),
        ("CalVer, Aura", "2025.01-aura", True),
    ],
)
@mock.patch(
    "neo4j_graphrag.experimental.components.kg_writer.Neo4jWriter._db_setup",
    return_value=None,
)
def test_get_version(
    _: Mock,
    driver: MagicMock,
    description: str,
    version: str,
    is_5_23_or_above: bool,
) -> None:
    execute_query_mock = MagicMock(
        return_value=(
            [
                {"versions": [version], "edition": "enterprise"},
            ],
            None,
            None,
        )
    )
    driver.execute_query = execute_query_mock
    neo4j_writer = Neo4jWriter(driver=driver)
    assert (
        neo4j_writer.is_version_5_23_or_above is is_5_23_or_above
    ), f"Failed is_version_5_23_or_above test case: {description}"



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Any
from unittest.mock import patch

import pytest

from neo4j_graphrag.exceptions import InvalidHybridSearchRankerError
from neo4j_graphrag.neo4j_queries import (
    get_query_tail,
    get_search_query,
    _get_hybrid_query_linear,
)
from neo4j_graphrag.types import EntityType, SearchType


def test_vector_search_basic() -> None:
    expected = (
        "CALL db.index.vector.queryNodes($vector_index_name, $top_k * $effective_search_ratio, $query_vector) "
        "YIELD node, score "
        "WITH node, score LIMIT $top_k "
        "RETURN node { .*, `None`: null } AS node, labels(node) AS nodeLabels, elementId(node) AS elementId, elementId(node) AS id, score"
    )
    result, params = get_search_query(SearchType.VECTOR)
    assert result.strip() == expected.strip()
    assert params == {}


def test_rel_vector_search_basic() -> None:
    expected = (
        "CALL db.index.vector.queryRelationships($vector_index_name, $top_k * $effective_search_ratio, $query_vector) "
        "YIELD relationship, score "
        "WITH relationship, score LIMIT $top_k "
        "RETURN relationship { .*, `None`: null } AS relationship, type(relationship) as relationshipType, "
        "elementId(relationship) AS elementId, elementId(relationship) AS id, score"
    )
    result, params = get_search_query(SearchType.VECTOR, EntityType.RELATIONSHIP)
    assert result.strip() == expected.strip()
    assert params == {}


def test_rel_vector_search_filters_err() -> None:
    with pytest.raises(Exception) as exc_info:
        get_search_query(
            SearchType.VECTOR, EntityType.RELATIONSHIP, filters={"filter": None}
        )
    assert str(exc_info.value) == "Filters are not supported for relationship indexes"


def test_rel_vector_search_hybrid_err() -> None:
    with pytest.raises(Exception) as exc_info:
        get_search_query(SearchType.HYBRID, EntityType.RELATIONSHIP)
    assert (
        str(exc_info.value) == "Hybrid search is not supported for relationship indexes"
    )


def test_hybrid_search_basic() -> None:
    expected = (
        "CALL { "
        "CALL db.index.vector.queryNodes($vector_index_name, $top_k * $effective_search_ratio, $query_vector) "
        "YIELD node, score "
        "WITH node, score LIMIT $top_k "
        "WITH collect({node:node, score:score}) AS nodes, max(score) AS vector_index_max_score "
        "UNWIND nodes AS n "
        "RETURN n.node AS node, (n.score / vector_index_max_score) AS score UNION "
        "CALL db.index.fulltext.queryNodes($fulltext_index_name, $query_text, {limit: $top_k}) "
        "YIELD node, score "
        "WITH collect({node:node, score:score}) AS nodes, max(score) AS ft_index_max_score "
        "UNWIND nodes AS n "
        "RETURN n.node AS node, (n.score / ft_index_max_score) AS score "
        "} "
        "WITH node, max(score) AS score ORDER BY score DESC LIMIT $top_k "
        "RETURN node { .*, `None`: null } AS node, labels(node) AS nodeLabels, elementId(node) AS elementId, elementId(node) AS id, score"
    )
    result, _ = get_search_query(SearchType.HYBRID)
    assert result.strip() == expected.strip()


def test_vector_search_with_properties() -> None:
    properties = ["name", "age"]
    expected = (
        "CALL db.index.vector.queryNodes($vector_index_name, $top_k * $effective_search_ratio, $query_vector) "
        "YIELD node, score "
        "WITH node, score LIMIT $top_k "
        "RETURN node {.name, .age} AS node, labels(node) AS nodeLabels, elementId(node) AS elementId, elementId(node) AS id, score"
    )
    result, _ = get_search_query(SearchType.VECTOR, return_properties=properties)
    assert result.strip() == expected.strip()


def test_vector_search_with_retrieval_query() -> None:
    retrieval_query = "MATCH (n) RETURN n LIMIT 10"
    expected = (
        "CALL db.index.vector.queryNodes($vector_index_name, $top_k * $effective_search_ratio, $query_vector) "
        "YIELD node, score "
        "WITH node, score LIMIT $top_k " + retrieval_query
    )
    result, _ = get_search_query(SearchType.VECTOR, retrieval_query=retrieval_query)
    assert result.strip() == expected.strip()


@patch("neo4j_graphrag.neo4j_queries.get_metadata_filter", return_value=["True", {}])
def test_vector_search_with_filters(_mock: Any) -> None:
    expected = (
        "MATCH (node:`Label`) "
        "WHERE node.`vector` IS NOT NULL "
        "AND size(node.`vector`) = toInteger($embedding_dimension)"
        " AND (True) "
        "WITH node, "
        "vector.similarity.cosine(node.`vector`, $query_vector) AS score "
        "ORDER BY score DESC LIMIT $top_k "
        "RETURN node { .*, `vector`: null } AS node, labels(node) AS nodeLabels, elementId(node) AS elementId, elementId(node) AS id, score"
    )
    result, params = get_search_query(
        SearchType.VECTOR,
        node_label="Label",
        embedding_node_property="vector",
        embedding_dimension=1,
        filters={"field": "value"},
    )
    assert result.strip() == expected.strip()
    assert params == {"embedding_dimension": 1}


@patch(
    "neo4j_graphrag.neo4j_queries.get_metadata_filter",
    return_value=["True", {"param": "value"}],
)
def test_vector_search_with_params_from_filters(_mock: Any) -> None:
    expected = (
        "MATCH (node:`Label`) "
        "WHERE node.`vector` IS NOT NULL "
        "AND size(node.`vector`) = toInteger($embedding_dimension)"
        " AND (True) "
        "WITH node, "
        "vector.similarity.cosine(node.`vector`, $query_vector) AS score "
        "ORDER BY score DESC LIMIT $top_k "
        "RETURN node { .*, `vector`: null } AS node, labels(node) AS nodeLabels, elementId(node) AS elementId, elementId(node) AS id, score"
    )
    result, params = get_search_query(
        SearchType.VECTOR,
        node_label="Label",
        embedding_node_property="vector",
        embedding_dimension=1,
        filters={"field": "value"},
    )
    assert result.strip() == expected.strip()
    assert params == {"embedding_dimension": 1, "param": "value"}


def test_hybrid_search_with_retrieval_query() -> None:
    retrieval_query = "MATCH (n) RETURN n LIMIT 10"
    expected = (
        "CALL { "
        "CALL db.index.vector.queryNodes($vector_index_name, $top_k * $effective_search_ratio, $query_vector) "
        "YIELD node, score "
        "WITH node, score LIMIT $top_k "
        "WITH collect({node:node, score:score}) AS nodes, max(score) AS vector_index_max_score "
        "UNWIND nodes AS n "
        "RETURN n.node AS node, (n.score / vector_index_max_score) AS score UNION "
        "CALL db.index.fulltext.queryNodes($fulltext_index_name, $query_text, {limit: $top_k}) "
        "YIELD node, score "
        "WITH collect({node:node, score:score}) AS nodes, max(score) AS ft_index_max_score "
        "UNWIND nodes AS n "
        "RETURN n.node AS node, (n.score / ft_index_max_score) AS score "
        "} "
        "WITH node, max(score) AS score ORDER BY score DESC LIMIT $top_k "
        + retrieval_query
    )
    result, _ = get_search_query(SearchType.HYBRID, retrieval_query=retrieval_query)
    assert result.strip() == expected.strip()


def test_hybrid_search_with_properties() -> None:
    properties = ["name", "age"]
    expected = (
        "CALL { "
        "CALL db.index.vector.queryNodes($vector_index_name, $top_k * $effective_search_ratio, $query_vector) "
        "YIELD node, score "
        "WITH node, score LIMIT $top_k "
        "WITH collect({node:node, score:score}) AS nodes, max(score) AS vector_index_max_score "
        "UNWIND nodes AS n "
        "RETURN n.node AS node, (n.score / vector_index_max_score) AS score UNION "
        "CALL db.index.fulltext.queryNodes($fulltext_index_name, $query_text, {limit: $top_k}) "
        "YIELD node, score "
        "WITH collect({node:node, score:score}) AS nodes, max(score) AS ft_index_max_score "
        "UNWIND nodes AS n "
        "RETURN n.node AS node, (n.score / ft_index_max_score) AS score "
        "} "
        "WITH node, max(score) AS score ORDER BY score DESC LIMIT $top_k "
        "RETURN node {.name, .age} AS node, labels(node) AS nodeLabels, elementId(node) AS elementId, elementId(node) AS id, score"
    )
    result, _ = get_search_query(SearchType.HYBRID, return_properties=properties)
    assert result.strip() == expected.strip()


def test_get_query_tail_with_retrieval_query() -> None:
    retrieval_query = "MATCH (n) RETURN n LIMIT 10"
    expected = retrieval_query
    result = get_query_tail(retrieval_query=retrieval_query)
    assert result.strip() == expected.strip()


def test_get_query_tail_with_properties() -> None:
    properties = ["name", "age"]
    expected = "RETURN node {.name, .age} AS node, labels(node) AS nodeLabels, elementId(node) AS elementId, elementId(node) AS id, score"
    result = get_query_tail(return_properties=properties)
    assert result.strip() == expected.strip()


def test_get_query_tail_with_fallback() -> None:
    fallback = "HELLO"
    expected = fallback
    result = get_query_tail(fallback_return=fallback)
    assert result.strip() == expected.strip()


def test_get_query_tail_ordering_all() -> None:
    retrieval_query = "MATCH (n) RETURN n LIMIT 10"
    properties = ["name", "age"]
    fallback = "HELLO"

    expected = retrieval_query
    result = get_query_tail(
        retrieval_query=retrieval_query,
        return_properties=properties,
        fallback_return=fallback,
    )
    assert result.strip() == expected.strip()


def test_get_query_tail_ordering_no_retrieval_query() -> None:
    properties = ["name", "age"]
    fallback = "HELLO"

    expected = "RETURN node {.name, .age} AS node, labels(node) AS nodeLabels, elementId(node) AS elementId, elementId(node) AS id, score"
    result = get_query_tail(
        return_properties=properties,
        fallback_return=fallback,
    )
    assert result.strip() == expected.strip()


def test_get_hybrid_query_linear_with_alpha() -> None:
    query = _get_hybrid_query_linear(neo4j_version_is_5_23_or_above=True, alpha=0.7)
    vector_substr = "rawScore * $alpha"
    ft_substr = "rawScore * (1 - $alpha)"
    assert vector_substr in query
    assert ft_substr in query


def test_invalid_hybrid_search_ranker_error() -> None:
    with pytest.raises(InvalidHybridSearchRankerError):
        get_search_query(SearchType.HYBRID, ranker="invalid")



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging

import pytest
from neo4j import Driver
from neo4j_graphrag.embeddings.base import Embedder
from neo4j_graphrag.retrievers import (
    HybridCypherRetriever,
    HybridRetriever,
)
from neo4j_graphrag.types import RetrieverResult, RetrieverResultItem


@pytest.mark.usefixtures("setup_neo4j_for_retrieval")
def test_hybrid_retriever_search_text(
    driver: Driver, random_embedder: Embedder
) -> None:
    retriever = HybridRetriever(
        driver, "vector-index-name", "fulltext-index-name", random_embedder
    )

    top_k = 5
    effective_search_ratio = 2
    results = retriever.search(
        query_text="Find me a book about Fremen",
        top_k=top_k,
        effective_search_ratio=effective_search_ratio,
    )

    assert isinstance(results, RetrieverResult)
    assert len(results.items) == 5
    for result in results.items:
        assert isinstance(result, RetrieverResultItem)
        assert "'vectorProperty': None," in result.content


@pytest.mark.usefixtures("setup_neo4j_for_retrieval")
def test_hybrid_retriever_no_neo4j_deprecation_warning(
    driver: Driver, random_embedder: Embedder, caplog: pytest.LogCaptureFixture
) -> None:
    retriever = HybridRetriever(
        driver, "vector-index-name", "fulltext-index-name", random_embedder
    )

    top_k = 5
    effective_search_ratio = 2
    with caplog.at_level(logging.WARNING):
        retriever.search(
            query_text="Find me a book about Fremen",
            top_k=top_k,
            effective_search_ratio=effective_search_ratio,
        )

    for record in caplog.records:
        if (
            "Neo.ClientNotification.Statement.FeatureDeprecationWarning"
            in record.message
        ):
            assert False, f"Deprecation warning found in logs: {record.message}"


@pytest.mark.usefixtures("setup_neo4j_for_retrieval")
def test_hybrid_cypher_retriever_search_text(
    driver: Driver, random_embedder: Embedder
) -> None:
    retrieval_query = "MATCH (node)-[:AUTHORED_BY]->(author:Author) RETURN author.name"
    retriever = HybridCypherRetriever(
        driver,
        "vector-index-name",
        "fulltext-index-name",
        retrieval_query,
        random_embedder,
    )

    top_k = 5
    effective_search_ratio = 2
    results = retriever.search(
        query_text="Find me a book about Fremen",
        top_k=top_k,
        effective_search_ratio=effective_search_ratio,
    )

    assert isinstance(results, RetrieverResult)
    assert len(results.items) == 5
    for record in results.items:
        assert isinstance(record, RetrieverResultItem)
        assert "author.name" in record.content


@pytest.mark.usefixtures("setup_neo4j_for_retrieval")
def test_hybrid_retriever_search_vector(driver: Driver) -> None:
    retriever = HybridRetriever(
        driver,
        "vector-index-name",
        "fulltext-index-name",
    )

    top_k = 5
    effective_search_ratio = 2
    results = retriever.search(
        query_text="Find me a book about Fremen",
        query_vector=[1.0 for _ in range(1536)],
        top_k=top_k,
        effective_search_ratio=effective_search_ratio,
    )

    assert isinstance(results, RetrieverResult)
    assert len(results.items) == 5
    for result in results.items:
        assert isinstance(result, RetrieverResultItem)


@pytest.mark.usefixtures("setup_neo4j_for_retrieval")
def test_hybrid_cypher_retriever_search_vector(driver: Driver) -> None:
    retrieval_query = "MATCH (node)-[:AUTHORED_BY]->(author:Author) RETURN author.name"
    retriever = HybridCypherRetriever(
        driver,
        "vector-index-name",
        "fulltext-index-name",
        retrieval_query,
    )

    top_k = 5
    effective_search_ratio = 2
    results = retriever.search(
        query_text="Find me a book about Fremen",
        query_vector=[1.0 for _ in range(1536)],
        top_k=top_k,
        effective_search_ratio=effective_search_ratio,
    )

    assert isinstance(results, RetrieverResult)
    assert len(results.items) == 5
    for record in results.items:
        assert isinstance(record, RetrieverResultItem)
        assert "author.name" in record.content


@pytest.mark.usefixtures("setup_neo4j_for_retrieval")
def test_hybrid_retriever_return_properties(driver: Driver) -> None:
    properties = ["name", "age"]
    retriever = HybridRetriever(
        driver,
        "vector-index-name",
        "fulltext-index-name",
        return_properties=properties,
    )

    top_k = 5
    effective_search_ratio = 2
    results = retriever.search(
        query_text="Find me a book about Fremen",
        query_vector=[1.0 for _ in range(1536)],
        top_k=top_k,
        effective_search_ratio=effective_search_ratio,
    )

    assert isinstance(results, RetrieverResult)
    assert len(results.items) == 5
    for result in results.items:
        assert isinstance(result, RetrieverResultItem)


@pytest.mark.usefixtures("setup_neo4j_for_retrieval")
def test_hybrid_retriever_search_text_linear_ranker(
    driver: Driver, random_embedder: Embedder
) -> None:
    retriever = HybridRetriever(
        driver, "vector-index-name", "fulltext-index-name", random_embedder
    )

    top_k = 5
    effective_search_ratio = 2
    results = retriever.search(
        query_text="Find me a book about Fremen",
        top_k=top_k,
        effective_search_ratio=effective_search_ratio,
        ranker="linear",
        alpha=0.9,
    )

    assert isinstance(results, RetrieverResult)
    assert len(results.items) == 5
    for result in results.items:
        assert isinstance(result, RetrieverResultItem)



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from pydantic import validate_call

from neo4j_graphrag.experimental.components.text_splitters.base import TextSplitter
from neo4j_graphrag.experimental.components.types import TextChunk, TextChunks


def _adjust_chunk_start(text: str, approximate_start: int) -> int:
    """
    Shift the starting index backward if it lands in the middle of a word.
    If no whitespace is found, use the proposed start.

     Args:
        text (str): The text being split.
        approximate_start (int): The initial starting index of the chunk.

    Returns:
        int: The adjusted starting index, ensuring the chunk does not begin in the
             middle of a word if possible.
    """
    start = approximate_start
    if start > 0 and not text[start].isspace() and not text[start - 1].isspace():
        while start > 0 and not text[start - 1].isspace():
            start -= 1

        # fallback if no whitespace is found
        if start == 0 and not text[0].isspace():
            start = approximate_start
    return start


def _adjust_chunk_end(text: str, start: int, approximate_end: int) -> int:
    """
    Shift the ending index backward if it lands in the middle of a word.
    If no whitespace is found, use 'approximate_end'.

    Args:
        text (str): The full text being split.
        start (int): The adjusted starting index for this chunk.
        approximate_end (int): The initial end index.

    Returns:
        int: The adjusted ending index, ensuring the chunk does not end in the middle of
            a word if possible.
    """
    end = approximate_end
    if end < len(text):
        while end > start and not text[end].isspace() and not text[end - 1].isspace():
            end -= 1

        # fallback if no whitespace is found
        if end == start:
            end = approximate_end
    return end


class FixedSizeSplitter(TextSplitter):
    """Text splitter which splits the input text into fixed or approximate fixed size
       chunks with optional overlap.

    Args:
        chunk_size (int): The number of characters in each chunk.
        chunk_overlap (int): The number of characters from the previous chunk to overlap
                            with each chunk. Must be less than `chunk_size`.
        approximate (bool): If True, avoids splitting words in the middle at chunk
                            boundaries. Defaults to True.


    Example:

    .. code-block:: python

        from neo4j_graphrag.experimental.components.text_splitters.fixed_size_splitter import FixedSizeSplitter
        from neo4j_graphrag.experimental.pipeline import Pipeline

        pipeline = Pipeline()
        text_splitter = FixedSizeSplitter(chunk_size=4000, chunk_overlap=200, approximate=True)
        pipeline.add_component(text_splitter, "text_splitter")
    """

    @validate_call
    def __init__(
        self, chunk_size: int = 4000, chunk_overlap: int = 200, approximate: bool = True
    ) -> None:
        if chunk_size <= 0:
            raise ValueError("chunk_size must be strictly greater than 0")
        if chunk_overlap >= chunk_size:
            raise ValueError("chunk_overlap must be strictly less than chunk_size")
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.approximate = approximate

    @validate_call
    async def run(self, text: str) -> TextChunks:
        """Splits a piece of text into chunks.

        Args:
            text (str): The text to be split.

        Returns:
            TextChunks: A list of chunks.
        """
        chunks = []
        index = 0
        step = self.chunk_size - self.chunk_overlap
        text_length = len(text)
        approximate_start = 0
        skip_adjust_chunk_start = False
        end = 0

        while end < text_length:
            if self.approximate:
                start = (
                    approximate_start
                    if skip_adjust_chunk_start
                    else _adjust_chunk_start(text, approximate_start)
                )
                # adjust start and end to avoid cutting words in the middle
                approximate_end = min(start + self.chunk_size, text_length)
                end = _adjust_chunk_end(text, start, approximate_end)
                # when avoiding splitting words in the middle is not possible, revert to
                # initial chunk end and skip adjusting next chunk start
                skip_adjust_chunk_start = end == approximate_end
            else:
                # apply fixed size splitting with possibly words cut in half at chunk
                # boundaries
                start = approximate_start
                end = min(start + self.chunk_size, text_length)

            chunk_text = text[start:end]
            chunks.append(TextChunk(text=chunk_text, index=index))
            index += 1

            approximate_start = start + step

        return TextChunks(chunks=chunks)



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from unittest.mock import Mock, call, patch

from neo4j_graphrag.utils.logging import Prettifier, prettify


def test_prettifier_short_str() -> None:
    p = Prettifier()
    value = "ab"
    pretty_value = p._prettify_str(value)
    assert pretty_value == "ab"


def test_prettifier_long_str() -> None:
    p = Prettifier()
    value = "ab" * 200
    pretty_value = p._prettify_str(value)
    assert pretty_value == "ab" * 100 + "... (200 chars)"


@patch("neo4j_graphrag.utils.logging.os.environ")
def test_prettifier_str_custom_max_length(mock_env: Mock) -> None:
    mock_env.return_value = {"LOGGING__MAX_STRING_LENGTH": "1"}
    p = Prettifier()
    value = "abc" * 100
    pretty_value = p._prettify_str(value)
    assert pretty_value == "a" + "... (299 chars)"


def test_prettifier_short_list() -> None:
    p = Prettifier()
    value = list("abc")
    pretty_value = p._prettify_list(value)
    assert pretty_value == ["a", "b", "c"]


def test_prettifier_long_list() -> None:
    p = Prettifier()
    value = list("abc") * 10
    pretty_value = p._prettify_list(value)
    assert pretty_value == ["a", "b", "c", "a", "b", "... (25 items)"]


@patch("neo4j_graphrag.utils.logging.os.environ")
def test_prettifier_list_custom_max_length(mock_env: Mock) -> None:
    mock_env.return_value = {"LOGGING__MAX_LIST_LENGTH": "1"}
    p = Prettifier()
    value = list("abc") * 10
    pretty_value = p._prettify_list(value)
    assert pretty_value == ["a", "... (29 items)"]


def test_prettifier_list_nested() -> None:
    with patch.object(
        Prettifier, "_prettify_str", return_value="mocked string"
    ) as mock:
        p = Prettifier()
        value = ["abc" * 200] * 6
        pretty_value = p._prettify_list(value)
        mock.assert_has_calls([call("abc" * 200)] * p.max_list_length)
        assert pretty_value == ["mocked string"] * 5 + ["... (1 items)"]


def test_prettifier_dict_nested() -> None:
    with patch.object(
        Prettifier, "_prettify_str", return_value="mocked string"
    ) as mock_str:
        with patch.object(
            Prettifier, "_prettify_list", return_value=["mocked list"]
        ) as mock_list:
            p = Prettifier()
            value = {
                "key1": "string",
                "key2": ["a", "list"],
            }
            pretty_value = p._prettify_dict(value)
            mock_str.assert_has_calls([call("string")])
            mock_list.assert_has_calls(
                [
                    call(["a", "list"]),
                ]
            )
            assert pretty_value == {
                "key1": "mocked string",
                "key2": ["mocked list"],
            }


def test_prettify_function() -> None:
    assert prettify(
        {
            "key": {
                "key0.1": "ab" * 200,
                "key0.2": ["a"] * 10,
                "key0.3": {"key0.3.1": "a short string"},
            }
        }
    ) == {
        "key": {
            "key0.1": "ab" * 100 + "... (200 chars)",
            "key0.2": ["a"] * 5 + ["... (5 items)"],
            "key0.3": {"key0.3.1": "a short string"},
        }
    }



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import neo4j
from neo4j_graphrag.message_history import Neo4jMessageHistory
from neo4j_graphrag.types import LLMMessage


def test_neo4j_message_history_add_message(driver: neo4j.Driver) -> None:
    driver.execute_query(query_="MATCH (n) DETACH DELETE n;")
    message_history = Neo4jMessageHistory(session_id="123", driver=driver)
    message_history.add_message(
        LLMMessage(role="user", content="Hello"),
    )
    assert len(message_history.messages) == 1
    assert message_history.messages[0]["role"] == "user"
    assert message_history.messages[0]["content"] == "Hello"


def test_neo4j_message_history_add_messages(driver: neo4j.Driver) -> None:
    driver.execute_query(query_="MATCH (n) DETACH DELETE n;")
    message_history = Neo4jMessageHistory(session_id="123", driver=driver)
    message_history.add_messages(
        [
            LLMMessage(role="system", content="You are a helpful assistant."),
            LLMMessage(role="user", content="Hello"),
            LLMMessage(
                role="assistant",
                content="Hello, how may I help you today?",
            ),
            LLMMessage(role="user", content="I'd like to buy a new car."),
            LLMMessage(
                role="assistant",
                content="I'd be happy to help you find the perfect car.",
            ),
        ]
    )
    assert len(message_history.messages) == 5
    assert message_history.messages[0]["role"] == "system"
    assert message_history.messages[0]["content"] == "You are a helpful assistant."
    assert message_history.messages[1]["role"] == "user"
    assert message_history.messages[1]["content"] == "Hello"
    assert message_history.messages[2]["role"] == "assistant"
    assert message_history.messages[2]["content"] == "Hello, how may I help you today?"
    assert message_history.messages[3]["role"] == "user"
    assert message_history.messages[3]["content"] == "I'd like to buy a new car."
    assert message_history.messages[4]["role"] == "assistant"
    assert (
        message_history.messages[4]["content"]
        == "I'd be happy to help you find the perfect car."
    )


def test_neo4j_message_history_clear_messages(driver: neo4j.Driver) -> None:
    driver.execute_query(query_="MATCH (n) DETACH DELETE n;")
    message_history = Neo4jMessageHistory(session_id="123", driver=driver)
    message_history.add_messages(
        [
            LLMMessage(role="system", content="You are a helpful assistant."),
            LLMMessage(role="user", content="Hello"),
        ]
    )
    assert len(message_history.messages) == 2
    message_history.clear()
    assert len(message_history.messages) == 0
    # Test that the session node is not deleted
    results = driver.execute_query(
        query_="MATCH (s:`Session`) WHERE s.id = '123' RETURN s"
    )
    assert len(results.records) == 1
    assert results.records[0]["s"]["id"] == "123"
    assert list(results.records[0]["s"].labels) == ["Session"]


def test_neo4j_message_history_clear_session_and_messages(driver: neo4j.Driver) -> None:
    driver.execute_query(query_="MATCH (n) DETACH DELETE n;")
    message_history = Neo4jMessageHistory(session_id="123", driver=driver)
    message_history.add_messages(
        [
            LLMMessage(role="system", content="You are a helpful assistant."),
            LLMMessage(role="user", content="Hello"),
        ]
    )
    assert len(message_history.messages) == 2
    message_history.clear(delete_session_node=True)
    assert len(message_history.messages) == 0
    # Test that the session node is deleted
    results = driver.execute_query(
        query_="MATCH (s:`Session`) WHERE s.id = '123' RETURN s"
    )
    assert results.records == []


def test_neo4j_message_history_clear_no_messages(driver: neo4j.Driver) -> None:
    driver.execute_query(query_="MATCH (n) DETACH DELETE n;")
    message_history = Neo4jMessageHistory(session_id="123", driver=driver)
    message_history.clear(delete_session_node=True)
    results = driver.execute_query(
        query_="MATCH (s:`Session`) WHERE s.id = '123' RETURN s"
    )
    assert results.records == []


def test_neo4j_message_window_size(driver: neo4j.Driver) -> None:
    driver.execute_query(query_="MATCH (n) DETACH DELETE n;")
    message_history = Neo4jMessageHistory(session_id="123", driver=driver, window=1)
    message_history.add_messages(
        [
            LLMMessage(role="system", content="You are a helpful assistant."),
            LLMMessage(role="user", content="Hello"),
            LLMMessage(
                role="assistant",
                content="Hello, how may I help you today?",
            ),
            LLMMessage(role="user", content="I'd like to buy a new car."),
            LLMMessage(
                role="assistant",
                content="I'd be happy to help you find the perfect car.",
            ),
        ]
    )
    assert len(message_history.messages) == 1
    assert (
        message_history.messages[0]["content"]
        == "I'd be happy to help you find the perfect car."
    )
    assert message_history.messages[0]["role"] == "assistant"


def test_neo4j_message_history_session_id_unique(driver: neo4j.Driver) -> None:
    driver.execute_query(query_="MATCH (n) DETACH DELETE n;")
    Neo4jMessageHistory(session_id="123", driver=driver)
    Neo4jMessageHistory(session_id="123", driver=driver)

    results = driver.execute_query(
        query_="MATCH (s:`Session`) WHERE s.id = '123' RETURN s"
    )
    records = results.records
    assert len(records) == 1
    session = records[0]["s"]
    assert session.get("createdAt") is not None
    assert session.get("updatedAt") is not None



"""
Example showing how to use VertexAI tool calls with parameter extraction.
Both synchronous and asynchronous examples are provided.
"""

import asyncio
from typing import Optional

from dotenv import load_dotenv
from vertexai.generative_models import GenerationConfig

from neo4j_graphrag.llm import VertexAILLM
from neo4j_graphrag.llm.types import ToolCallResponse
from neo4j_graphrag.tool import (
    Tool,
    ObjectParameter,
    StringParameter,
    IntegerParameter,
)

# Load environment variables from .env file
load_dotenv()


# Create a custom Tool implementation for person info extraction
person_tool_parameters = ObjectParameter(
    description="Parameters for extracting person information",
    properties={
        "name": StringParameter(description="The person's full name"),
        "age": IntegerParameter(description="The person's age"),
        "occupation": StringParameter(description="The person's occupation"),
    },
    required_properties=["name"],
    additional_properties=False,
)


def run_person_tool(
    name: str, age: Optional[int] = None, occupation: Optional[str] = None
) -> str:
    """A simple function that summarizes person information from input parameters."""
    return f"Found person {name} with age {age} and occupation {occupation}"


person_info_tool = Tool(
    name="extract_person_info",
    description="Extract information about a person from text",
    parameters=person_tool_parameters,
    execute_func=run_person_tool,
)

company_tool_parameters = ObjectParameter(
    description="Parameters for extracting company information",
    properties={
        "name": StringParameter(description="The company's full name"),
        "industry": StringParameter(description="The company's industry"),
        "creation_year": IntegerParameter(description="The company's creation year"),
    },
    required_properties=["name"],
    additional_properties=False,
)


def run_company_tool(
    name: str, industry: Optional[str] = None, creation_year: Optional[int] = None
) -> str:
    """A simple function that summarizes company information from input parameters."""
    return (
        f"Found company {name} operating in industry {industry} since {creation_year}"
    )


company_info_tool = Tool(
    name="extract_company_info",
    description="Extract information about a company from text",
    parameters=company_tool_parameters,
    execute_func=run_company_tool,
)

# Create the tool instance
TOOLS = [person_info_tool, company_info_tool]


def process_tool_call(response: ToolCallResponse) -> str:
    """Process the tool call response and return the extracted parameters."""
    if not response.tool_calls:
        raise ValueError("No tool calls found in response")

    tool_call = response.tool_calls[0]
    print(f"\nTool called: {tool_call.name}")
    print(f"Arguments: {tool_call.arguments}")
    print(f"Additional content: {response.content or 'None'}")
    if tool_call.name == "extract_person_info":
        return person_info_tool.execute(**tool_call.arguments)  # type: ignore[no-any-return]
    elif tool_call.name == "extract_company_info":
        return str(company_info_tool.execute(**tool_call.arguments))
    else:
        raise ValueError("Unknown tool call")


async def main() -> None:
    # Initialize the VertexAI LLM
    generation_config = GenerationConfig(temperature=0.0)
    llm = VertexAILLM(
        model_name="gemini-2.0-flash-001",
        generation_config=generation_config,
        # tool_config=ToolConfig(
        #     function_calling_config=ToolConfig.FunctionCallingConfig(
        #         mode=ToolConfig.FunctionCallingConfig.Mode.ANY,
        #         # allowed_function_names=["extract_person_info"],
        #     ))
    )

    # Example text containing information about a company
    text1 = "Neo4j is a software company created in 2007"

    print("\n=== Synchronous Tool Call ===")
    # Make a synchronous tool call
    sync_response = llm.invoke_with_tools(
        input=f"Extract information about the person from this text: {text1}",
        tools=TOOLS,
    )
    sync_result = process_tool_call(sync_response)
    print("\n=== Synchronous Tool Call Result ===")
    print(sync_result)

    print("\n=== Asynchronous Tool Call ===")
    # Make an asynchronous tool call with a different text about a person
    text2 = "Molly Hane, 32, works as a data scientist and enjoys machine learning."
    async_response = await llm.ainvoke_with_tools(
        input=f"Extract information about the person from this text: {text2}",
        tools=TOOLS,
    )
    async_result = process_tool_call(async_response)
    print("\n=== Asynchronous Tool Call Result ===")
    print(async_result)


if __name__ == "__main__":
    # Run the async main function
    asyncio.run(main())



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

import logging
from typing import Any, Callable, Optional, Union

import neo4j
from pydantic import ValidationError

from neo4j_graphrag.embeddings.base import Embedder
from neo4j_graphrag.exceptions import (
    EmbeddingRequiredError,
    RetrieverInitializationError,
    SearchValidationError,
    SearchQueryParseError,
)
from neo4j_graphrag.neo4j_queries import get_search_query
from neo4j_graphrag.retrievers.base import Retriever
from neo4j_graphrag.types import (
    EmbedderModel,
    HybridCypherRetrieverModel,
    HybridCypherSearchModel,
    HybridRetrieverModel,
    HybridSearchModel,
    Neo4jDriverModel,
    RawSearchResult,
    RetrieverResultItem,
    SearchType,
    HybridSearchRanker,
)
from neo4j_graphrag.utils.logging import prettify

logger = logging.getLogger(__name__)


class HybridRetriever(Retriever):
    """
    Provides retrieval method using combination of vector search over embeddings and
    fulltext search.
    If an embedder is provided, it needs to have the required Embedder type.

    Example:

    .. code-block:: python

      import neo4j
      from neo4j_graphrag.retrievers import HybridRetriever

      driver = neo4j.GraphDatabase.driver(URI, auth=AUTH)

      retriever = HybridRetriever(
          driver, "vector-index-name", "fulltext-index-name", custom_embedder
      )
      retriever.search(query_text="Find me a book about Fremen", top_k=5)

    Args:
        driver (neo4j.Driver): The Neo4j Python driver.
        vector_index_name (str): Vector index name.
        fulltext_index_name (str): Fulltext index name.
        embedder (Optional[Embedder]): Embedder object to embed query text.
        return_properties (Optional[list[str]]): List of node properties to return.
        result_formatter (Optional[Callable[[neo4j.Record], RetrieverResultItem]]): Provided custom function to transform a neo4j.Record to a RetrieverResultItem.
        neo4j_database (Optional[str]): The name of the Neo4j database. If not provided, this defaults to the server's default database ("neo4j" by default) (`see reference to documentation <https://neo4j.com/docs/operations-manual/current/database-administration/#manage-databases-default>`_).

            Two variables are provided in the neo4j.Record:

            -   node: Represents the node retrieved from the vector index search.
            -   score: Denotes the similarity score.
    """

    def __init__(
        self,
        driver: neo4j.Driver,
        vector_index_name: str,
        fulltext_index_name: str,
        embedder: Optional[Embedder] = None,
        return_properties: Optional[list[str]] = None,
        result_formatter: Optional[
            Callable[[neo4j.Record], RetrieverResultItem]
        ] = None,
        neo4j_database: Optional[str] = None,
    ) -> None:
        try:
            driver_model = Neo4jDriverModel(driver=driver)
            embedder_model = EmbedderModel(embedder=embedder) if embedder else None
            validated_data = HybridRetrieverModel(
                driver_model=driver_model,
                vector_index_name=vector_index_name,
                fulltext_index_name=fulltext_index_name,
                embedder_model=embedder_model,
                return_properties=return_properties,
                result_formatter=result_formatter,
                neo4j_database=neo4j_database,
            )
        except ValidationError as e:
            raise RetrieverInitializationError(e.errors()) from e

        super().__init__(
            validated_data.driver_model.driver, validated_data.neo4j_database
        )
        self.vector_index_name = validated_data.vector_index_name
        self.fulltext_index_name = validated_data.fulltext_index_name
        self.return_properties = validated_data.return_properties
        self.embedder = (
            validated_data.embedder_model.embedder
            if validated_data.embedder_model
            else None
        )
        self.result_formatter = validated_data.result_formatter
        self._embedding_node_property = None
        self._embedding_dimension = None
        self._fetch_index_infos(self.vector_index_name)

    def default_record_formatter(self, record: neo4j.Record) -> RetrieverResultItem:
        """
        Best effort to guess the node-to-text method. Inherited classes
        can override this method to implement custom text formatting.
        """
        metadata = {
            "score": record.get("score"),
        }
        node = record.get("node")
        return RetrieverResultItem(
            content=str(node),
            metadata=metadata,
        )

    def get_search_results(
        self,
        query_text: str,
        query_vector: Optional[list[float]] = None,
        top_k: int = 5,
        effective_search_ratio: int = 1,
        ranker: Union[str, HybridSearchRanker] = HybridSearchRanker.NAIVE,
        alpha: Optional[float] = None,
    ) -> RawSearchResult:
        """Get the top_k nearest neighbor embeddings for either provided query_vector or query_text.
        Both query_vector and query_text can be provided.
        If query_vector is provided, then it will be preferred over the embedded query_text
        for the vector search.

        See the following documentation for more details:

        - `Query a vector index <https://neo4j.com/docs/cypher-manual/current/indexes-for-vector-search/#indexes-vector-query>`_
        - `db.index.vector.queryNodes() <https://neo4j.com/docs/operations-manual/5/reference/procedures/#procedure_db_index_vector_queryNodes>`_
        - `db.index.fulltext.queryNodes() <https://neo4j.com/docs/operations-manual/5/reference/procedures/#procedure_db_index_fulltext_querynodes>`_

        To query by text, an embedder must be provided when the class is instantiated.

        Args:
            query_text (str): The text to get the closest neighbors of.
            query_vector (Optional[list[float]], optional): The vector embeddings to get the closest neighbors of. Defaults to None.
            top_k (int, optional): The number of neighbors to return. Defaults to 5.
            effective_search_ratio (int): Controls the candidate pool size for the vector index by multiplying top_k to balance query
                accuracy and performance. Defaults to 1.
            ranker (str, HybridSearchRanker): Type of ranker to order the results from retrieval.
            alpha (Optional[float]): Weight for the vector score when using the linear ranker.
                The fulltext index score is multiplied by (1 - alpha).
                **Required** when using the linear ranker; must be between 0 and 1.

        Raises:
            SearchValidationError: If validation of the input arguments fail.
            EmbeddingRequiredError: If no embedder is provided.

        Returns:
            RawSearchResult: The results of the search query as a list of neo4j.Record and an optional metadata dict
        """
        try:
            validated_data = HybridSearchModel(
                query_vector=query_vector,
                query_text=query_text,
                top_k=top_k,
                effective_search_ratio=effective_search_ratio,
                ranker=ranker,
                alpha=alpha,
            )
        except ValidationError as e:
            raise SearchValidationError(e.errors()) from e

        parameters = validated_data.model_dump(exclude_none=True)
        parameters["vector_index_name"] = self.vector_index_name
        parameters["fulltext_index_name"] = self.fulltext_index_name

        if query_text and not query_vector:
            if not self.embedder:
                raise EmbeddingRequiredError(
                    "Embedding method required for text query."
                )
            query_vector = self.embedder.embed_query(query_text)
            parameters["query_vector"] = query_vector
        search_query, _ = get_search_query(
            search_type=SearchType.HYBRID,
            return_properties=self.return_properties,
            embedding_node_property=self._embedding_node_property,
            neo4j_version_is_5_23_or_above=self.neo4j_version_is_5_23_or_above,
            ranker=validated_data.ranker,
            alpha=validated_data.alpha,
        )

        if "ranker" in parameters:
            del parameters["ranker"]

        logger.debug("HybridRetriever Cypher parameters: %s", prettify(parameters))
        logger.debug("HybridRetriever Cypher query: %s", search_query)

        try:
            records, _, _ = self.driver.execute_query(
                search_query,
                parameters,
                database_=self.neo4j_database,
                routing_=neo4j.RoutingControl.READ,
            )
        except neo4j.exceptions.ClientError as e:
            if "org.apache.lucene.queryparser.classic.ParseException" in str(e):
                raise SearchQueryParseError(
                    f"Invalid Lucene query generated from query_text: {query_text}"
                ) from e
            raise
        return RawSearchResult(
            records=records,
            metadata={"query_vector": query_vector},
        )


class HybridCypherRetriever(Retriever):
    """
    Provides retrieval method using combination of vector search over embeddings and
    fulltext search, augmented by a Cypher query.
    This retriever builds on HybridRetriever.
    If an embedder is provided, it needs to have the required Embedder type.

    Note: `node` is a variable from the base query that can be used in `retrieval_query` as seen in the example below.

    Example:

    .. code-block:: python

      import neo4j
      from neo4j_graphrag.retrievers import HybridCypherRetriever

      driver = neo4j.GraphDatabase.driver(URI, auth=AUTH)

      retrieval_query = "MATCH (node)-[:AUTHORED_BY]->(author:Author)" "RETURN author.name"
      retriever = HybridCypherRetriever(
          driver, "vector-index-name", "fulltext-index-name", retrieval_query, custom_embedder
      )
      retriever.search(query_text="Find me a book about Fremen", top_k=5)

    To query by text, an embedder must be provided when the class is instantiated.

    Args:
        driver (neo4j.Driver): The Neo4j Python driver.
        vector_index_name (str): Vector index name.
        fulltext_index_name (str): Fulltext index name.
        retrieval_query (str): Cypher query that gets appended.
        embedder (Optional[Embedder]): Embedder object to embed query text.
        result_formatter (Optional[Callable[[neo4j.Record], RetrieverResultItem]]): Provided custom function to transform a neo4j.Record to a RetrieverResultItem.
        neo4j_database (Optional[str]): The name of the Neo4j database. If not provided, this defaults to the server's default database ("neo4j" by default) (`see reference to documentation <https://neo4j.com/docs/operations-manual/current/database-administration/#manage-databases-default>`_).

    Raises:
        RetrieverInitializationError: If validation of the input arguments fail.
    """

    def __init__(
        self,
        driver: neo4j.Driver,
        vector_index_name: str,
        fulltext_index_name: str,
        retrieval_query: str,
        embedder: Optional[Embedder] = None,
        result_formatter: Optional[
            Callable[[neo4j.Record], RetrieverResultItem]
        ] = None,
        neo4j_database: Optional[str] = None,
    ) -> None:
        try:
            driver_model = Neo4jDriverModel(driver=driver)
            embedder_model = EmbedderModel(embedder=embedder) if embedder else None
            validated_data = HybridCypherRetrieverModel(
                driver_model=driver_model,
                vector_index_name=vector_index_name,
                fulltext_index_name=fulltext_index_name,
                retrieval_query=retrieval_query,
                embedder_model=embedder_model,
                result_formatter=result_formatter,
                neo4j_database=neo4j_database,
            )
        except ValidationError as e:
            raise RetrieverInitializationError(e.errors()) from e

        super().__init__(
            validated_data.driver_model.driver, validated_data.neo4j_database
        )
        self.vector_index_name = validated_data.vector_index_name
        self.fulltext_index_name = validated_data.fulltext_index_name
        self.retrieval_query = validated_data.retrieval_query
        self.embedder = (
            validated_data.embedder_model.embedder
            if validated_data.embedder_model
            else None
        )
        self.result_formatter = validated_data.result_formatter

    def get_search_results(
        self,
        query_text: str,
        query_vector: Optional[list[float]] = None,
        top_k: int = 5,
        effective_search_ratio: int = 1,
        query_params: Optional[dict[str, Any]] = None,
        ranker: Union[str, HybridSearchRanker] = HybridSearchRanker.NAIVE,
        alpha: Optional[float] = None,
    ) -> RawSearchResult:
        """Get the top_k nearest neighbor embeddings for either provided query_vector or query_text.
        Both query_vector and query_text can be provided.
        If query_vector is provided, then it will be preferred over the embedded query_text
        for the vector search.

        See the following documentation for more details:

        - `Query a vector index <https://neo4j.com/docs/cypher-manual/current/indexes-for-vector-search/#indexes-vector-query>`_
        - `db.index.vector.queryNodes() <https://neo4j.com/docs/operations-manual/5/reference/procedures/#procedure_db_index_vector_queryNodes>`_
        - `db.index.fulltext.queryNodes() <https://neo4j.com/docs/operations-manual/5/reference/procedures/#procedure_db_index_fulltext_querynodes>`_

        Args:
            query_text (str): The text to get the closest neighbors of.
            query_vector (Optional[list[float]]): The vector embeddings to get the closest neighbors of. Defaults to None.
            top_k (int): The number of neighbors to return. Defaults to 5.
            effective_search_ratio (int): Controls the candidate pool size for the vector index by multiplying top_k to balance query
                accuracy and performance. Defaults to 1.
            query_params (Optional[dict[str, Any]]): Parameters for the Cypher query. Defaults to None.
            ranker (str, HybridSearchRanker): Type of ranker to order the results from retrieval.
            alpha (Optional[float]): Weight for the vector score when using the linear ranker.
                The fulltext index score is multiplied by (1 - alpha).
                **Required** when using the linear ranker; must be between 0 and 1.
        Raises:
            SearchValidationError: If validation of the input arguments fail.
            EmbeddingRequiredError: If no embedder is provided.

        Returns:
            RawSearchResult: The results of the search query as a list of neo4j.Record and an optional metadata dict
        """
        try:
            validated_data = HybridCypherSearchModel(
                query_vector=query_vector,
                query_text=query_text,
                top_k=top_k,
                effective_search_ratio=effective_search_ratio,
                ranker=ranker,
                alpha=alpha,
                query_params=query_params,
            )
        except ValidationError as e:
            raise SearchValidationError(e.errors()) from e

        parameters = validated_data.model_dump(exclude_none=True)
        parameters["vector_index_name"] = self.vector_index_name
        parameters["fulltext_index_name"] = self.fulltext_index_name

        if query_text and not query_vector:
            if not self.embedder:
                raise EmbeddingRequiredError(
                    "Embedding method required for text query."
                )
            query_vector = self.embedder.embed_query(query_text)
            parameters["query_vector"] = query_vector

        if query_params:
            for key, value in query_params.items():
                if key not in parameters:
                    parameters[key] = value
            del parameters["query_params"]

        search_query, _ = get_search_query(
            search_type=SearchType.HYBRID,
            retrieval_query=self.retrieval_query,
            neo4j_version_is_5_23_or_above=self.neo4j_version_is_5_23_or_above,
            ranker=validated_data.ranker,
            alpha=validated_data.alpha,
        )

        if "ranker" in parameters:
            del parameters["ranker"]

        logger.debug("HybridRetriever Cypher parameters: %s", prettify(parameters))
        logger.debug("HybridRetriever Cypher query: %s", search_query)

        try:
            records, _, _ = self.driver.execute_query(
                search_query,
                parameters,
                database_=self.neo4j_database,
                routing_=neo4j.RoutingControl.READ,
            )
        except neo4j.exceptions.ClientError as e:
            if "org.apache.lucene.queryparser.classic.ParseException" in str(e):
                raise SearchQueryParseError(
                    f"Invalid Lucene query generated from query_text: {query_text}"
                ) from e
            raise
        return RawSearchResult(
            records=records,
            metadata={"query_vector": query_vector},
        )



"""This example illustrates how to get started easily with the SimpleKGPipeline
and ingest PDF into a Neo4j Knowledge Graph.

This example assumes a Neo4j db is up and running. Update the credentials below
if needed.

OPENAI_API_KEY needs to be in the env vars.
"""

import asyncio
from pathlib import Path

import neo4j
from neo4j_graphrag.embeddings import OpenAIEmbeddings
from neo4j_graphrag.experimental.pipeline.kg_builder import SimpleKGPipeline
from neo4j_graphrag.experimental.pipeline.pipeline import PipelineResult
from neo4j_graphrag.llm import LLMInterface
from neo4j_graphrag.llm import OpenAILLM

# Neo4j db infos
URI = "neo4j://localhost:7687"
AUTH = ("neo4j", "password")
DATABASE = "neo4j"


root_dir = Path(__file__).parents[1]
file_path = root_dir / "data" / "Harry Potter and the Chamber of Secrets Summary.pdf"


# Instantiate NodeType and RelationshipType objects. This defines the
# entities and relations the LLM will be looking for in the text.
NODE_TYPES = ["Person", "Organization", "Location"]
RELATIONSHIP_TYPES = ["SITUATED_AT", "INTERACTS", "LED_BY"]
PATTERNS = [
    ("Person", "SITUATED_AT", "Location"),
    ("Person", "INTERACTS", "Person"),
    ("Organization", "LED_BY", "Person"),
]


async def define_and_run_pipeline(
    neo4j_driver: neo4j.Driver,
    llm: LLMInterface,
) -> PipelineResult:
    # Create an instance of the SimpleKGPipeline
    kg_builder = SimpleKGPipeline(
        llm=llm,
        driver=neo4j_driver,
        embedder=OpenAIEmbeddings(),
        schema={
            "node_types": NODE_TYPES,
            "relationship_types": RELATIONSHIP_TYPES,
            "patterns": PATTERNS,
        },
        neo4j_database=DATABASE,
    )
    return await kg_builder.run_async(
        file_path=str(file_path),
        # optional, add document metadata, each item will
        # be saved as a property of the Document node
        # document_metadata={"author": "J. K. Rowling"},
    )


async def main() -> PipelineResult:
    llm = OpenAILLM(
        model_name="gpt-4o",
        model_params={
            "max_tokens": 2000,
            "response_format": {"type": "json_object"},
        },
    )
    with neo4j.GraphDatabase.driver(URI, auth=AUTH) as driver:
        res = await define_and_run_pipeline(driver, llm)
    await llm.async_client.close()
    return res


if __name__ == "__main__":
    res = asyncio.run(main())
    print(res)



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

from typing import Any

import neo4j
from neo4j import GraphDatabase
from pinecone import Pinecone, ServerlessSpec

from ..utils import build_data_objects, populate_neo4j


def populate_dbs(
    neo4j_driver: neo4j.Driver, pc_client: Pinecone, index_name: str = "jeopardy"
) -> None:
    neo4j_objects, pc_question_objs = build_data_objects("pinecone")

    pc_client.create_index(
        name=index_name,
        dimension=384,
        metric="cosine",
        spec=ServerlessSpec(cloud="aws", region="us-east-1"),
    )

    # Populate Pinecone
    populate_pinecone(pc_client, pc_question_objs, index_name)

    # Populate Neo4j
    populate_neo4j(neo4j_driver, neo4j_objects)


def populate_pinecone(
    pc_client: Pinecone, pc_question_objs: list[Any], index_name: str
) -> None:
    index = pc_client.Index(index_name)
    index.upsert(vectors=pc_question_objs)


if __name__ == "__main__":
    NEO4J_AUTH = ("neo4j", "password")
    NEO4J_URL = "neo4j://localhost:7687"
    PC_API_KEY = "API_KEY"
    with GraphDatabase.driver(NEO4J_URL, auth=NEO4J_AUTH) as neo4j_driver:
        pc_client = Pinecone(PC_API_KEY)
        populate_dbs(neo4j_driver, pc_client)



"""This example illustrates how to get started easily with the SimpleKGPipeline
and ingest text into a Neo4j Knowledge Graph.

This example assumes a Neo4j db is up and running. Update the credentials below
if needed.

NB: when building a KG from text, no 'Document' node is created in the Knowledge Graph.
"""

import asyncio
import logging

import neo4j
from neo4j_graphrag.embeddings import OpenAIEmbeddings
from neo4j_graphrag.experimental.pipeline.kg_builder import SimpleKGPipeline
from neo4j_graphrag.experimental.pipeline.pipeline import PipelineResult
from neo4j_graphrag.experimental.pipeline.types.schema import (
    EntityInputType,
    RelationInputType,
)
from neo4j_graphrag.llm import LLMInterface
from neo4j_graphrag.llm import OpenAILLM

logging.basicConfig()
logging.getLogger("neo4j_graphrag").setLevel(logging.DEBUG)
# logging.getLogger("neo4j_graphrag").setLevel(logging.INFO)


# Neo4j db infos
URI = "neo4j://localhost:7687"
AUTH = ("neo4j", "password")
DATABASE = "neo4j"

# Text to process
TEXT = """The son of Duke Leto Atreides and the Lady Jessica, Paul is the heir of House Atreides,
an aristocratic family that rules the planet Caladan, the rainy planet, since 10191."""

# Instantiate Entity and Relation objects. This defines the
# entities and relations the LLM will be looking for in the text.
NODE_TYPES: list[EntityInputType] = [
    # entities can be defined with a simple label...
    "Person",
    # ... or with a dict if more details are needed,
    # such as a description:
    {"label": "House", "description": "Family the person belongs to"},
    # or a list of properties the LLM will try to attach to the entity:
    {"label": "Planet", "properties": [{"name": "weather", "type": "STRING"}]},
]
# same thing for relationships:
RELATIONSHIP_TYPES: list[RelationInputType] = [
    "PARENT_OF",
    {
        "label": "HEIR_OF",
        "description": "Used for inheritor relationship between father and sons",
    },
    {"label": "RULES", "properties": [{"name": "fromYear", "type": "INTEGER"}]},
]
PATTERNS = [
    ("Person", "PARENT_OF", "Person"),
    ("Person", "HEIR_OF", "House"),
    ("House", "RULES", "Planet"),
]


async def define_and_run_pipeline(
    neo4j_driver: neo4j.Driver,
    llm: LLMInterface,
) -> PipelineResult:
    # Create an instance of the SimpleKGPipeline
    kg_builder = SimpleKGPipeline(
        llm=llm,
        driver=neo4j_driver,
        embedder=OpenAIEmbeddings(),
        schema={
            "node_types": NODE_TYPES,
            "relationship_types": RELATIONSHIP_TYPES,
            "patterns": PATTERNS,
        },
        from_pdf=False,
        neo4j_database=DATABASE,
    )
    return await kg_builder.run_async(
        text=TEXT,
        # optional, specify file path for the Document node
        # if not, a random name will be generated
        # file_path="my_document.txt"
        # optional, add document metadata, each item will
        # be saved as a property of the Document node
        # document_metadata={"author": "Frank Herbert"},
    )


async def main() -> PipelineResult:
    llm = OpenAILLM(
        model_name="gpt-4o",
        model_params={
            "max_tokens": 2000,
            "response_format": {"type": "json_object"},
        },
    )
    with neo4j.GraphDatabase.driver(URI, auth=AUTH) as driver:
        res = await define_and_run_pipeline(driver, llm)
    await llm.async_client.close()
    return res


if __name__ == "__main__":
    res = asyncio.run(main())
    print(res)



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from unittest import mock
from unittest.mock import MagicMock, call

import pytest
from neo4j_graphrag.exceptions import RagInitializationError, SearchValidationError
from neo4j_graphrag.generation.graphrag import GraphRAG
from neo4j_graphrag.generation.prompts import RagTemplate
from neo4j_graphrag.generation.types import RagResultModel
from neo4j_graphrag.llm import LLMResponse, LLMInterfaceV2
from neo4j_graphrag.message_history import InMemoryMessageHistory
from neo4j_graphrag.types import LLMMessage, RetrieverResult, RetrieverResultItem


@pytest.fixture(scope="function")
def llm_v2() -> MagicMock:
    return MagicMock(spec=LLMInterfaceV2)


def test_graphrag_prompt_template() -> None:
    template = RagTemplate()
    prompt = template.format(
        context="my context", query_text="user's query", examples=""
    )
    assert (
        prompt
        == """Context:
my context

Examples:


Question:
user's query

Answer:
"""
    )


def test_graphrag_happy_path(retriever_mock: MagicMock, llm_v2: MagicMock) -> None:
    rag = GraphRAG(
        retriever=retriever_mock,
        llm=llm_v2,
    )
    retriever_mock.search.return_value = RetrieverResult(
        items=[
            RetrieverResultItem(content="item content 1"),
            RetrieverResultItem(content="item content 2"),
        ]
    )
    llm_v2.invoke.return_value = LLMResponse(content="llm generated text")

    res = rag.search("question", retriever_config={"top_k": 111})

    retriever_mock.search.assert_called_once_with(query_text="question", top_k=111)
    llm_v2.invoke.assert_called_once_with(
        input=[
            {
                "role": "system",
                "content": "Answer the user question using the provided context.",
            },
            {
                "role": "user",
                "content": """Context:
item content 1
item content 2

Examples:


Question:
question

Answer:
""",
            },
        ],
    )

    assert isinstance(res, RagResultModel)
    assert res.answer == "llm generated text"
    assert res.retriever_result is None


def test_graphrag_happy_path_with_message_history(
    retriever_mock: MagicMock, llm_v2: MagicMock
) -> None:
    rag = GraphRAG(
        retriever=retriever_mock,
        llm=llm_v2,
    )
    retriever_mock.search.return_value = RetrieverResult(
        items=[
            RetrieverResultItem(content="item content 1"),
            RetrieverResultItem(content="item content 2"),
        ]
    )
    llm_v2.invoke.side_effect = [
        LLMResponse(content="llm generated summary"),
        LLMResponse(content="llm generated text"),
    ]
    message_history = [
        {"role": "user", "content": "initial question"},
        {"role": "assistant", "content": "answer to initial question"},
    ]
    res = rag.search("question", message_history)  # type: ignore

    expected_retriever_query_text = """
Message Summary:
llm generated summary

Current Query:
question
"""

    first_invocation_input = """
Summarize the message history:

user: initial question
assistant: answer to initial question
"""
    first_invocation_system_instruction = "You are a summarization assistant. Summarize the given text in no more than 300 words."
    second_invocation = """Context:
item content 1
item content 2

Examples:


Question:
question

Answer:
"""

    retriever_mock.search.assert_called_once_with(
        query_text=expected_retriever_query_text
    )
    assert llm_v2.invoke.call_count == 2
    llm_v2.invoke.assert_has_calls(
        [
            # First call for summarization uses V2 interface
            call(
                input=[
                    {
                        "role": "system",
                        "content": first_invocation_system_instruction,
                    },
                    {"role": "user", "content": first_invocation_input},
                ],
            ),
            # Second call uses V2 interface
            call(
                input=[
                    {
                        "role": "system",
                        "content": "Answer the user question using the provided context.",
                    },
                    {"role": "user", "content": "initial question"},
                    {"role": "assistant", "content": "answer to initial question"},
                    {"role": "user", "content": second_invocation},
                ],
            ),
        ]
    )

    assert isinstance(res, RagResultModel)
    assert res.answer == "llm generated text"
    assert res.retriever_result is None


def test_graphrag_happy_path_with_in_memory_message_history(
    retriever_mock: MagicMock, llm_v2: MagicMock
) -> None:
    rag = GraphRAG(
        retriever=retriever_mock,
        llm=llm_v2,
    )
    retriever_mock.search.return_value = RetrieverResult(
        items=[
            RetrieverResultItem(content="item content 1"),
            RetrieverResultItem(content="item content 2"),
        ]
    )
    llm_v2.invoke.side_effect = [
        LLMResponse(content="llm generated summary"),
        LLMResponse(content="llm generated text"),
    ]
    message_history = InMemoryMessageHistory(
        messages=[
            LLMMessage(role="user", content="initial question"),
            LLMMessage(role="assistant", content="answer to initial question"),
        ]
    )
    res = rag.search("question", message_history)

    expected_retriever_query_text = """
Message Summary:
llm generated summary

Current Query:
question
"""

    first_invocation_input = """
Summarize the message history:

user: initial question
assistant: answer to initial question
"""
    first_invocation_system_instruction = "You are a summarization assistant. Summarize the given text in no more than 300 words."
    second_invocation = """Context:
item content 1
item content 2

Examples:


Question:
question

Answer:
"""

    retriever_mock.search.assert_called_once_with(
        query_text=expected_retriever_query_text
    )
    assert llm_v2.invoke.call_count == 2
    llm_v2.invoke.assert_has_calls(
        [
            # First call for summarization uses V2 interface
            call(
                input=[
                    {
                        "role": "system",
                        "content": first_invocation_system_instruction,
                    },
                    {"role": "user", "content": first_invocation_input},
                ],
            ),
            # Second call uses V2 interface
            call(
                input=[
                    {
                        "role": "system",
                        "content": "Answer the user question using the provided context.",
                    },
                    {"role": "user", "content": "initial question"},
                    {"role": "assistant", "content": "answer to initial question"},
                    {"role": "user", "content": second_invocation},
                ],
            ),
        ]
    )

    assert isinstance(res, RagResultModel)
    assert res.answer == "llm generated text"
    assert res.retriever_result is None


def test_graphrag_happy_path_custom_system_instruction(
    retriever_mock: MagicMock, llm_v2: MagicMock
) -> None:
    prompt_template = RagTemplate(system_instructions="Custom instruction")
    rag = GraphRAG(
        retriever=retriever_mock,
        llm=llm_v2,
        prompt_template=prompt_template,
    )
    retriever_mock.search.return_value = RetrieverResult(items=[])
    llm_v2.invoke.side_effect = [
        LLMResponse(content="llm generated text"),
    ]
    res = rag.search("question")

    assert llm_v2.invoke.call_count == 1
    llm_v2.invoke.assert_has_calls(
        [
            call(
                input=[
                    {"role": "system", "content": "Custom instruction"},
                    {"role": "user", "content": mock.ANY},
                ],
            ),
        ]
    )

    assert res.answer == "llm generated text"


def test_graphrag_happy_path_response_fallback(
    retriever_mock: MagicMock, llm_v2: MagicMock
) -> None:
    rag = GraphRAG(
        retriever=retriever_mock,
        llm=llm_v2,
    )
    retriever_mock.search.return_value = RetrieverResult(items=[])
    res = rag.search(
        "question",
        response_fallback="I can't answer this question without context",
    )

    assert llm_v2.invoke.call_count == 0
    assert res.answer == "I can't answer this question without context"


def test_graphrag_initialization_error(llm_v2: MagicMock) -> None:
    with pytest.raises(RagInitializationError) as excinfo:
        GraphRAG(
            retriever="not a retriever object",  # type: ignore
            llm=llm_v2,
        )
    assert "retriever" in str(excinfo)


def test_graphrag_search_error(retriever_mock: MagicMock, llm_v2: MagicMock) -> None:
    rag = GraphRAG(
        retriever=retriever_mock,
        llm=llm_v2,
    )
    with pytest.raises(SearchValidationError) as excinfo:
        rag.search(10)  # type: ignore
    assert "Input should be a valid string" in str(excinfo)


def test_chat_summary_template(retriever_mock: MagicMock, llm_v2: MagicMock) -> None:
    message_history = [
        {"role": "user", "content": "initial question"},
        {"role": "assistant", "content": "answer to initial question"},
        {"role": "user", "content": "second question"},
        {"role": "assistant", "content": "answer to second question"},
    ]
    rag = GraphRAG(
        retriever=retriever_mock,
        llm=llm_v2,
    )
    prompt = rag._chat_summary_prompt(message_history=message_history)  # type: ignore
    assert (
        prompt
        == """
Summarize the message history:

user: initial question
assistant: answer to initial question
user: second question
assistant: answer to second question
"""
    )


def test_conversation_template(retriever_mock: MagicMock, llm_v2: MagicMock) -> None:
    rag = GraphRAG(
        retriever=retriever_mock,
        llm=llm_v2,
    )
    prompt = rag.conversation_prompt(
        summary="llm generated chat summary", current_query="latest question"
    )
    assert (
        prompt
        == """
Message Summary:
llm generated chat summary

Current Query:
latest question
"""
    )



import neo4j
from neo4j_graphrag.experimental.components.kg_writer import (
    KGWriterModel,
    Neo4jWriter,
)
from neo4j_graphrag.experimental.components.types import Neo4jGraph


async def main(driver: neo4j.Driver, graph: Neo4jGraph) -> KGWriterModel:
    writer = Neo4jWriter(
        driver,
        # optionally, configure the neo4j database
        # neo4j_database="neo4j",
        # you can tune batch_size to
        # improve speed
        # batch_size=1000,
    )
    result = await writer.run(graph=graph)
    return result



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from unittest.mock import MagicMock

import pytest
from neo4j_graphrag.message_history import InMemoryMessageHistory, Neo4jMessageHistory
from neo4j_graphrag.types import LLMMessage
from pydantic import ValidationError


def test_in_memory_message_history_add_message() -> None:
    message_history = InMemoryMessageHistory()
    message_history.add_message(
        LLMMessage(role="user", content="may thy knife chip and shatter")
    )
    assert len(message_history.messages) == 1
    assert message_history.messages[0]["role"] == "user"
    assert message_history.messages[0]["content"] == "may thy knife chip and shatter"


def test_in_memory_message_history_add_messages() -> None:
    message_history = InMemoryMessageHistory()
    message_history.add_messages(
        [
            LLMMessage(role="user", content="may thy knife chip and shatter"),
            LLMMessage(
                role="assistant",
                content="He who controls the spice controls the universe.",
            ),
        ]
    )
    assert len(message_history.messages) == 2
    assert message_history.messages[0]["role"] == "user"
    assert message_history.messages[0]["content"] == "may thy knife chip and shatter"
    assert message_history.messages[1]["role"] == "assistant"
    assert (
        message_history.messages[1]["content"]
        == "He who controls the spice controls the universe."
    )


def test_in_memory_message_history_clear() -> None:
    message_history = InMemoryMessageHistory()
    message_history.add_messages(
        [
            LLMMessage(role="user", content="may thy knife chip and shatter"),
            LLMMessage(
                role="assistant",
                content="He who controls the spice controls the universe.",
            ),
        ]
    )
    assert len(message_history.messages) == 2
    message_history.clear()
    assert len(message_history.messages) == 0


def test_neo4j_message_history_invalid_session_id(driver: MagicMock) -> None:
    with pytest.raises(ValidationError) as exc_info:
        Neo4jMessageHistory(session_id=1.5, driver=driver, window=1)  # type: ignore[arg-type]
    assert "Input should be a valid string" in str(exc_info.value)


def test_neo4j_message_history_invalid_driver() -> None:
    with pytest.raises(ValidationError) as exc_info:
        Neo4jMessageHistory(session_id="123", driver=1.5, window=1)  # type: ignore[arg-type]
    assert "Input should be an instance of Driver" in str(exc_info.value)


def test_neo4j_message_history_invalid_window(driver: MagicMock) -> None:
    with pytest.raises(ValidationError) as exc_info:
        Neo4jMessageHistory(session_id="123", driver=driver, window=-1)
    assert "Input should be greater than 0" in str(exc_info.value)


def test_neo4j_message_history_messages_setter(driver: MagicMock) -> None:
    message_history = Neo4jMessageHistory(session_id="123", driver=driver)
    with pytest.raises(NotImplementedError) as exc_info:
        message_history.messages = [
            LLMMessage(role="user", content="may thy knife chip and shatter"),
        ]
    assert (
        str(exc_info.value)
        == "Direct assignment to 'messages' is not allowed. Use the 'add_messages' instead."
    )


def test_neo4j_message_history_messages_getter_custom_db(driver: MagicMock) -> None:
    driver.execute_query.side_effect = [
        MagicMock(records=[]),
        MagicMock(
            records=[{"result": {"data": {"content": "my message"}, "role": "user"}}]
        ),
    ]
    message_history = Neo4jMessageHistory(
        session_id="123", driver=driver, database="my_db"
    )
    messages = message_history.messages
    assert len(messages) == 1
    assert messages[0] == LLMMessage(content="my message", role="user")
    for c in driver.execute_query.call_args_list:
        assert c.kwargs["database_"] == "my_db"



"""This example demonstrates how to use WeaviateNeo4jRetriever, ie vectors are
stored in the Weaviate database.

See the [README](./README.md) for more
information about how spin up a Weaviate and Neo4j databases if needed.

In this example, search is performed from an already existing vector.
"""

from embedding_biology import EMBEDDING_BIOLOGY
from neo4j import GraphDatabase
from neo4j_graphrag.retrievers import WeaviateNeo4jRetriever
from weaviate.connect.helpers import connect_to_local

NEO4J_URL = "neo4j://localhost:7687"
NEO4J_AUTH = ("neo4j", "password")


def main() -> None:
    with GraphDatabase.driver(NEO4J_URL, auth=NEO4J_AUTH) as neo4j_driver:
        with connect_to_local() as w_client:
            retriever = WeaviateNeo4jRetriever(
                driver=neo4j_driver,
                client=w_client,
                collection="Jeopardy",
                id_property_external="neo4j_id",
                id_property_neo4j="id",
            )
            res = retriever.search(query_vector=EMBEDDING_BIOLOGY, top_k=2)
            print(res)


if __name__ == "__main__":
    main()



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from unittest.mock import MagicMock, Mock, patch

import pytest
from neo4j_graphrag.embeddings.ollama import OllamaEmbeddings
from neo4j_graphrag.exceptions import EmbeddingsGenerationError


@patch("builtins.__import__", side_effect=ImportError)
def test_ollama_embedder_missing_dependency(mock_import: Mock) -> None:
    with pytest.raises(ImportError):
        OllamaEmbeddings(model="test")


@patch("builtins.__import__")
def test_ollama_embedder_happy_path(mock_import: Mock) -> None:
    mock_import.return_value.Client.return_value.embed.return_value = MagicMock(
        embeddings=[[1.0, 2.0]],
    )
    embedder = OllamaEmbeddings(model="test")
    res = embedder.embed_query("my text")
    assert isinstance(res, list)
    assert res == [1.0, 2.0]


@patch("builtins.__import__")
def test_ollama_embedder_empty_list(mock_import: Mock) -> None:
    mock_import.return_value.Client.return_value.embed.return_value = MagicMock(
        embeddings=[],
    )
    embedder = OllamaEmbeddings(model="test")
    with pytest.raises(EmbeddingsGenerationError):
        embedder.embed_query("my text")



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

import weaviate.classes as wvc
from neo4j import Driver, GraphDatabase
from weaviate.client import WeaviateClient
from weaviate.collections.classes.types import WeaviateField
from weaviate.connect.helpers import connect_to_local

from ..utils import build_data_objects, populate_neo4j


def populate_dbs(
    neo4j_driver: Driver, w_client: WeaviateClient, collection_name: str = "Jeopardy"
) -> None:
    neo4j_objects, w_question_objs = build_data_objects("weaviate")
    w_client.collections.create(
        collection_name,
        vectorizer_config=wvc.config.Configure.Vectorizer.text2vec_transformers(),
        vector_index_config=wvc.config.Configure.VectorIndex.hnsw(
            distance_metric=wvc.config.VectorDistances.COSINE  # select preferred distance metric
        ),
        properties=[
            wvc.config.Property(name="neo4j_id", data_type=wvc.config.DataType.TEXT),
        ],
    )

    # Populate Weaviate
    populate_weaviate(w_client, w_question_objs, collection_name)

    # Populate Neo4j
    populate_neo4j(neo4j_driver, neo4j_objects)


def populate_weaviate(
    w_client: WeaviateClient,
    w_question_objs: list[wvc.data.DataObject[dict[str, WeaviateField]]],
    collection_name: str,
) -> None:
    questions = w_client.collections.get(collection_name)
    questions.data.insert_many(w_question_objs)


if __name__ == "__main__":
    neo4j_auth = ("neo4j", "password")
    neo4j_url = "neo4j://localhost:7687"
    with GraphDatabase.driver(neo4j_url, auth=neo4j_auth) as neo4j_driver:
        with connect_to_local() as w_client:
            populate_dbs(neo4j_driver, w_client)
